from ..engine import Engine, SingleCoreEngine
from ..objective import HistoryOptions
from ..store.save_to_hdf5 import get_or_create_group
from pathlib import Path
from typing import Union


import h5py


def check_hdf5_mp(
    history_options: HistoryOptions,
    engine: Engine,
) -> Union[str, None]:
    """
    Create a folder for partial HDF5 files,
    if a parallelization engine will be used.

    Parameters
    ----------
    engine:
        The Engine which is used in the optimization
    history_options:
        The HistoryOptions used in the optimization

    Returns
    -------
    The filename that will be used to combine the partial HDF5 files later.
    If a parallelization engine is not used, `None` is returned.
    """
    if isinstance(engine, SingleCoreEngine):
        return None
    filename = history_options.storage_file
    file_path = Path(filename)

    # create directory with same name as original file stem
    partial_file_path = (
            file_path.parent / file_path.stem /
            (file_path.stem + '_{id}' + file_path.suffix)
    )
    partial_file_path.parent.mkdir(parents=True, exist_ok=True)
    history_options.storage_file = str(partial_file_path)

    # create hdf5 file that gathers the others within history group
    with h5py.File(filename, mode='a') as f:
        get_or_create_group(f, "history")
    return filename


def fill_hdf5_file(
    ret: list,
    filename: str
) -> None:
    """
    Create links in `filename` to the
    history of each start contained in ret, the results
    of the optimization.

    Parameters
    ----------
    ret:
        The result iterable returned by the optimization.
    filename:
        The filename of the hdf5 file in which the histories
        are to be gathered.
    """
    with h5py.File(filename, mode='a') as f:
        for result in ret:
            id = result['id']
            f[f'history/{id}'] = h5py.ExternalLink(
                result['history'].file,
                f'history/{id}'
            )



import numpy as np
import scipy.optimize
import re
import abc
import time
import logging
from typing import Dict, Optional

from ..objective import (
    HistoryOptions, HistoryBase, OptimizerHistory, CsvHistory, Hdf5History,
)
from ..problem import Problem
from ..objective.constants import MODE_FUN, MODE_RES, FVAL, GRAD
from .result import OptimizerResult

try:
    import cyipopt
except ImportError:
    cyipopt = None

try:
    import dlib
except ImportError:
    dlib = None

try:
    import pyswarm
except ImportError:
    pyswarm = None

try:
    import cma
except ImportError:
    cma = None

try:
    import pyswarms
except ImportError:
    pyswarms = None

try:
    import nlopt
except ImportError:
    nlopt = None

try:
    import fides
    from fides.hessian_approximation import HessianApproximation
except ImportError:
    fides = None
    HessianApproximation = None


EXITFLAG_LOADED_FROM_FILE = -99

logger = logging.getLogger(__name__)


def history_decorator(minimize):
    """
    Default decorator for the minimize() method to initialize and extract
    information stored in the history.
    """

    def wrapped_minimize(self, problem, x0, id, allow_failed_starts,
                         history_options=None):
        objective = problem.objective

        # initialize the objective
        objective.initialize()

        # create optimizer history
        if history_options is None:
            history_options = HistoryOptions()
        history = history_options.create_history(
            id=id, x_names=[problem.x_names[ix]
                            for ix in problem.x_free_indices])
        optimizer_history = OptimizerHistory(history=history, x0=x0)

        # plug in history for the objective to record it
        objective.history = optimizer_history

        # perform the actual minimization
        try:
            result = minimize(self, problem, x0, id, history_options)
            objective.history.finalize()
            result.id = id
        except Exception as err:
            if allow_failed_starts:
                logger.error(f'start {id} failed: {err}')
                result = OptimizerResult(
                    x0=x0, exitflag=-1, message=str(err), id=id)
            else:
                raise

        result = fill_result_from_objective_history(result, objective.history)

        # clean up, history is available from result
        objective.history = HistoryBase()

        return result

    return wrapped_minimize


def time_decorator(minimize):
    """
    Default decorator for the minimize() method to take time.
    Currently, the method time.time() is used, which measures
    the wall-clock time.
    """

    def wrapped_minimize(self, problem, x0, id, allow_failed_starts,
                         history_options=None):
        start_time = time.time()
        result = minimize(self, problem, x0, id, allow_failed_starts,
                          history_options)
        used_time = time.time() - start_time
        result.time = used_time
        return result

    return wrapped_minimize


def fix_decorator(minimize):
    """
    Default decorator for the minimize() method to include also fixed
    parameters in the result arrays (nans will be inserted in the
    derivatives).
    """

    def wrapped_minimize(self, problem, x0, id, allow_failed_starts,
                         history_options=None):
        # perform the actual optimization
        result = minimize(self, problem, x0, id, allow_failed_starts,
                          history_options)

        # vectors to full vectors
        result.update_to_full(problem)

        logger.info(f"Final fval={result.fval:.4f}, "
                    f"time={result.time:.4f}s, "
                    f"n_fval={result.n_fval}.")

        return result

    return wrapped_minimize


def fill_result_from_objective_history(
        result: OptimizerResult,
        optimizer_history: OptimizerHistory):
    """
    Overwrite function values in the result object with the values recorded in
    the history.
    """
    update_vals = True
    # check history for better values
    # value could be different e.g. if constraints violated
    if optimizer_history.fval_min is not None and result.fval is not None \
            and not np.isclose(optimizer_history.fval_min, result.fval):
        logger.warning(
            "Function values from history and optimizer do not match: "
            f"{optimizer_history.fval_min}, {result.fval}")
        # do not use value from history
        update_vals = False

    if optimizer_history.x_min is not None and result.x is not None \
            and not np.allclose(result.x, optimizer_history.x_min):
        logger.warning(
            "Parameters obtained from history and optimizer do not match: "
            f"{optimizer_history.x_min}, {result.x}")
        # do not use value from history
        update_vals = result.fval is None

    # update optimal values from history if reported function values
    # and parameters coincide
    if update_vals:
        # override values from history if available
        result.x = optimizer_history.x_min
        result.fval = optimizer_history.fval_min
        if optimizer_history.grad_min is not None:
            result.grad = optimizer_history.grad_min
        if optimizer_history.hess_min is not None:
            result.hess = optimizer_history.hess_min
        if optimizer_history.res_min is not None:
            result.res = optimizer_history.res_min
        if optimizer_history.sres_min is not None:
            result.sres = optimizer_history.sres_min

    # initial values
    result.x0 = optimizer_history.x0
    result.fval0 = optimizer_history.fval0

    # counters
    # we only use our own counters here as optimizers may report differently
    result.n_fval = optimizer_history.history.n_fval
    result.n_grad = optimizer_history.history.n_grad
    result.n_hess = optimizer_history.history.n_hess
    result.n_res = optimizer_history.history.n_res
    result.n_sres = optimizer_history.history.n_sres

    # trace
    result.history = optimizer_history.history

    return result


def read_result_from_file(problem: Problem, history_options: HistoryOptions,
                          identifier: str):
    """Fill an OptimizerResult from history."""
    if history_options.storage_file.endswith('.csv'):
        history = CsvHistory(
            file=history_options.storage_file.format(id=identifier),
            options=history_options,
            load_from_file=True
        )
    elif history_options.storage_file.endswith(('.h5', '.hdf5')):
        history = Hdf5History.load(
            id=identifier,
            file=history_options.storage_file.format(id=identifier),
        )
    else:
        raise NotImplementedError()

    opt_hist = OptimizerHistory(
        history, history.get_x_trace(0),
        generate_from_history=True
    )

    result = OptimizerResult(
        id=identifier,
        message='loaded from file',
        exitflag=EXITFLAG_LOADED_FROM_FILE,
        time=max(history.get_time_trace())
    )
    result.id = identifier
    result = fill_result_from_objective_history(result, opt_hist)
    result.update_to_full(problem)

    return result


class Optimizer(abc.ABC):
    """
    This is the optimizer base class, not functional on its own.
    An optimizer takes a problem, and possibly a start point, and then
    performs an optimization. It returns an OptimizerResult.
    """

    def __init__(self):
        """
        Default constructor.
        """

    @abc.abstractmethod
    @fix_decorator
    @time_decorator
    @history_decorator
    def minimize(
            self,
            problem: Problem,
            x0: np.ndarray,
            id: str,
            history_options: HistoryOptions = None,
    ) -> OptimizerResult:
        """"
        Perform optimization.
        Parameters
        ----------
        problem:
            The problem to find optimal parameters for.
        x0:
            The starting parameters.
        id:
            Multistart id.
        history_options:
            Optimizer history options.
        """

    @abc.abstractmethod
    def is_least_squares(self):
        return False

    def get_default_options(self):
        """
        Create default options specific for the optimizer.
        """
        return None


def check_finite_bounds(lb, ub):
    """Raise if bounds are not finite."""
    if not np.isfinite(lb).all() or not np.isfinite(ub).all():
        raise ValueError('Selected optimizer cannot work with unconstrained '
                         'optimization problems.')


class ScipyOptimizer(Optimizer):
    """
    Use the SciPy optimizers.
    Find details on the optimizer and configuration options at:
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.\
        optimize.minimize.html#scipy.optimize.minimize
    """

    def __init__(self,
                 method: str = 'L-BFGS-B',
                 tol: float = 1e-9,
                 options: Dict = None):
        super().__init__()

        self.method = method

        self.options = options
        if self.options is None:
            self.options = ScipyOptimizer.get_default_options(self)
            self.options['ftol'] = tol
        elif self.options is not None and 'ftol' not in self.options:
            self.options['ftol'] = tol

    @fix_decorator
    @time_decorator
    @history_decorator
    def minimize(
        self,
        problem: Problem,
        x0: np.ndarray,
        id: str,
        history_options: HistoryOptions = None,
    ) -> OptimizerResult:
        lb = problem.lb
        ub = problem.ub
        objective = problem.objective

        if self.is_least_squares():
            # is a residual based least squares method

            if not objective.has_res:
                raise Exception(
                    "For least squares optimization, the objective "
                    "must be able to compute residuals.")

            ls_method = self.method[3:]
            bounds = (lb, ub)

            fun = objective.get_res
            jac = objective.get_sres if objective.has_sres else '2-point'
            # TODO: pass jac computing methods in options

            if self.options is not None:
                ls_options = self.options.copy()
                ls_options['verbose'] = 2 if 'disp' in ls_options.keys() \
                                             and ls_options['disp'] else 0
                ls_options.pop('disp', None)
                ls_options['max_nfev'] = ls_options.pop('maxiter', None)
            else:
                ls_options = {}

            # optimize
            res = scipy.optimize.least_squares(
                fun=fun,
                x0=x0,
                method=ls_method,
                jac=jac,
                bounds=bounds,
                tr_solver=ls_options.pop('tr_solver',
                                         'lsmr' if len(x0) > 1 else 'exact'),
                loss='linear',
                **ls_options
            )
            # extract fval/grad from result, note that fval is not available
            # from least squares solvers
            grad = getattr(res, 'grad', None)
            fval = None
        else:
            # is an fval based optimization method

            if not objective.has_fun:
                raise Exception("For this optimizer, the objective must "
                                "be able to compute function values")

            bounds = scipy.optimize.Bounds(lb, ub)

            # fun_may_return_tuple = self.method.lower() in \
            #    ['cg', 'bfgs', 'newton-cg', 'l-bfgs-b', 'tnc', 'slsqp',
            #        'dogleg', 'trust-ncg']
            # TODO: is it more efficient to have tuple as output of fun?
            method_supports_grad = self.method.lower() in \
                ['cg', 'bfgs', 'newton-cg', 'l-bfgs-b', 'tnc', 'slsqp',
                 'dogleg', 'trust-ncg', 'trust-krylov', 'trust-exact',
                 'trust-constr']
            method_supports_hess = self.method.lower() in \
                ['newton-cg', 'dogleg', 'trust-ncg', 'trust-krylov',
                 'trust-exact', 'trust-constr']
            method_supports_hessp = self.method.lower() in \
                ['newton-cg', 'trust-ncg', 'trust-krylov', 'trust-constr']

            fun = objective.get_fval
            jac = objective.get_grad \
                if objective.has_grad and method_supports_grad \
                else None
            hess = objective.get_hess \
                if objective.has_hess and method_supports_hess \
                else None
            hessp = objective.get_hessp \
                if objective.has_hessp and method_supports_hessp \
                else None
            # minimize will ignore hessp otherwise
            if hessp is not None:
                hess = None

            # optimize
            res = scipy.optimize.minimize(
                fun=fun,
                x0=x0,
                method=self.method,
                jac=jac,
                hess=hess,
                hessp=hessp,
                bounds=bounds,
                options=self.options,
            )
            # extract fval/grad from result
            grad = getattr(res, 'jac', None)
            fval = res.fun

        # fill in everything known, although some parts will be overwritten
        optimizer_result = OptimizerResult(
            x=np.array(res.x),
            fval=fval,
            grad=grad,
            hess=getattr(res, 'hess', None),
            exitflag=res.status,
            message=res.message
        )

        return optimizer_result

    def is_least_squares(self):
        return re.match(r'(?i)^(ls_)', self.method)

    def get_default_options(self):
        if self.is_least_squares():
            options = {'max_nfev': 1000, 'disp': False}
        else:
            options = {'maxiter': 1000, 'disp': False}
        return options


class IpoptOptimizer(Optimizer):
    """Use IpOpt (https://pypi.org/project/ipopt/) for optimization."""

    def __init__(
            self, options: Dict = None):
        """
        Parameters
        ----------
        options:
            Options are directly passed on to `cyipopt.minimize_ipopt`.
        """
        super().__init__()
        self.options = options

    @fix_decorator
    @time_decorator
    @history_decorator
    def minimize(
            self,
            problem: Problem,
            x0: np.ndarray,
            id: str,
            history_options: HistoryOptions = None,
    ) -> OptimizerResult:

        if cyipopt is None:
            raise ImportError(
                "This optimizer requires an installation of ipopt. You can "
                "install ipopt via `pip install ipopt`."
            )

        objective = problem.objective

        bounds = np.array([problem.lb, problem.ub]).T

        ret = cyipopt.minimize_ipopt(
            fun=objective.get_fval,
            x0=x0,
            method=None,  # ipopt does not use this argument for anything
            jac=objective.get_grad,
            hess=None,  # ipopt does not support Hessian yet
            hessp=None,  # ipopt does not support Hessian vector product yet
            bounds=bounds,
            tol=None,  # can be set via options
            options=self.options,
        )

        # the ipopt return object is a scipy.optimize.OptimizeResult
        return OptimizerResult(
            x=ret.x,
            exitflag=ret.status,
            message=ret.message
        )

    def is_least_squares(self):
        return False


class DlibOptimizer(Optimizer):
    """
    Use the Dlib toolbox for optimization.
    """

    def __init__(self,
                 options: Dict = None):
        super().__init__()

        self.options = options
        if self.options is None:
            self.options = DlibOptimizer.get_default_options(self)
        elif 'maxiter' not in self.options:
            raise KeyError('Dlib options are missing the key word '
                           'maxiter.')

    @fix_decorator
    @time_decorator
    @history_decorator
    def minimize(
            self,
            problem: Problem,
            x0: np.ndarray,
            id: str,
            history_options: HistoryOptions = None,
    ) -> OptimizerResult:

        lb = problem.lb
        ub = problem.ub
        check_finite_bounds(lb, ub)
        objective = problem.objective

        if dlib is None:
            raise ImportError(
                "This optimizer requires an installation of dlib. You can "
                "install dlib via `pip install dlib`."
            )

        if not objective.has_fun:
            raise ValueError("For this optimizer, the objective must "
                             "be able to return function values.")

        # dlib requires variable length arguments
        def get_fval_vararg(*x):
            return objective.get_fval(x)

        dlib.find_min_global(
            get_fval_vararg,
            list(lb),
            list(ub),
            int(self.options['maxiter']),
            0.002,
        )

        optimizer_result = OptimizerResult()

        return optimizer_result

    def is_least_squares(self):
        return False

    def get_default_options(self):
        return {'maxiter': 10000}


class PyswarmOptimizer(Optimizer):
    """
    Global optimization using pyswarm.
    """

    def __init__(self, options: Dict = None):
        super().__init__()

        if options is None:
            options = {'maxiter': 200}
        self.options = options

    @fix_decorator
    @time_decorator
    @history_decorator
    def minimize(
            self,
            problem: Problem,
            x0: np.ndarray,
            id: str,
            history_options: HistoryOptions = None,
    ) -> OptimizerResult:
        lb = problem.lb
        ub = problem.ub
        if pyswarm is None:
            raise ImportError(
                "This optimizer requires an installation of pyswarm. You can "
                "install pyswarm via `pip install pyswarm."
            )

        check_finite_bounds(lb, ub)

        xopt, fopt = pyswarm.pso(
            problem.objective.get_fval, lb, ub, **self.options)

        optimizer_result = OptimizerResult(
            x=np.array(xopt),
            fval=fopt
        )

        return optimizer_result

    def is_least_squares(self):
        return False


class CmaesOptimizer(Optimizer):
    """
    Global optimization using cma-es.
    Package homepage: https://pypi.org/project/cma-es/
    """

    def __init__(self, par_sigma0: float = 0.25, options: Dict = None):
        """
        Parameters
        ----------
        par_sigma0:
            scalar, initial standard deviation in each coordinate.
            par_sigma0 should be about 1/4th of the search domain width
            (where the optimum is to be expected)
        options:
            Optimizer options that are directly passed on to cma.
        """

        super().__init__()

        if options is None:
            options = {'maxiter': 10000}
        self.options = options
        self.par_sigma0 = par_sigma0

    @fix_decorator
    @time_decorator
    @history_decorator
    def minimize(
            self,
            problem: Problem,
            x0: np.ndarray,
            id: str,
            history_options: HistoryOptions = None,
    ) -> OptimizerResult:

        lb = problem.lb
        ub = problem.ub

        check_finite_bounds(lb, ub)

        sigma0 = self.par_sigma0 * np.median(ub - lb)
        self.options['bounds'] = [lb, ub]

        if cma is None:
            raise ImportError(
                "This optimizer requires an installation of cma. You can "
                "install cma via `pip install cma."
            )

        result = cma.CMAEvolutionStrategy(
            x0, sigma0, inopts=self.options,
        ).optimize(problem.objective.get_fval).result

        optimizer_result = OptimizerResult(x=np.array(result[0]),
                                           fval=result[1])

        return optimizer_result

    def is_least_squares(self):
        return False


class ScipyDifferentialEvolutionOptimizer(Optimizer):
    """
    Global optimization using scipy's differential evolution optimizer.
    Package homepage: https://docs.scipy.org/doc/scipy/reference/generated\
        /scipy.optimize.differential_evolution.html

    Parameters
    ----------
    options:
        Optimizer options that are directly passed on to scipy's optimizer.


    Examples
    --------
    Arguments that can be passed to options:

    maxiter:
        used to calculate the maximal number of funcion evaluations by
        maxfevals = (maxiter + 1) * popsize * len(x)
        Default: 100
    popsize:
        population size, default value 15
    """

    def __init__(self, options: Dict = None):
        super().__init__()

        if options is None:
            options = {'maxiter': 100}
        self.options = options

    @fix_decorator
    @time_decorator
    @history_decorator
    def minimize(
            self,
            problem: Problem,
            x0: np.ndarray,
            id: str,
            history_options: HistoryOptions = None,
    ) -> OptimizerResult:
        bounds = list(zip(problem.lb, problem.ub))

        result = scipy.optimize.differential_evolution(
            problem.objective.get_fval, bounds, **self.options
        )

        optimizer_result = OptimizerResult(x=np.array(result.x),
                                           fval=result.fun)

        return optimizer_result

    def is_least_squares(self):
        return False


class PyswarmsOptimizer(Optimizer):
    """
    Global optimization using pyswarms.
    Package homepage: https://pyswarms.readthedocs.io/en/latest/index.html

    Parameters
    ----------
    par_popsize:
        number of particles in the swarm, default value 10

    options:
        Optimizer options that are directly passed on to pyswarms.
        c1: cognitive parameter
        c2: social parameter
        w: inertia parameter
        Default values are (c1,c2,w) = (0.5, 0.3, 0.9)

    Examples
    --------
    Arguments that can be passed to options:

    maxiter:
        used to calculate the maximal number of funcion evaluations.
        Default: 1000
    """

    def __init__(self, par_popsize: float = 10, options: Dict = None):
        super().__init__()

        all_options = {'maxiter': 1000, 'c1': 0.5, 'c2': 0.3, 'w': 0.9}
        if options is None:
            options = {}
        all_options.update(options)
        self.options = all_options
        self.par_popsize = par_popsize

    @fix_decorator
    @time_decorator
    @history_decorator
    def minimize(
            self,
            problem: Problem,
            x0: np.ndarray,
            id: str,
            history_options: HistoryOptions = None,
    ) -> OptimizerResult:
        lb = problem.lb
        ub = problem.ub

        if pyswarms is None:
            raise ImportError(
                "This optimizer requires an installation of pyswarms.")

        # check for finite values for the bounds
        if np.isfinite(lb).all() is np.False_:
            raise ValueError(
                "This optimizer can only handle finite lower bounds.")
        if np.isfinite(ub).all() is np.False_:
            raise ValueError(
                "This optimizer can only handle finite upper bounds.")

        optimizer = pyswarms.single.global_best.GlobalBestPSO(
            n_particles=self.par_popsize, dimensions=len(x0),
            options=self.options, bounds=(lb, ub))

        def successively_working_fval(swarm: np.ndarray) -> np.ndarray:
            """Evaluate the function for all parameters in the swarm object.

            Parameters:
            -----------
            swarm: np.ndarray, shape (n_particales_in_swarm, n_parameters)

            Returns:
            --------
            result: np.ndarray, shape (n_particles_in_swarm)
            """

            n_particles = swarm.shape[0]
            result = np.zeros(n_particles)
            # iterate over the particles in the swarm
            for i_particle, par in enumerate(swarm):
                result[i_particle] = problem.objective.get_fval(par)

            return result

        cost, pos = optimizer.optimize(
            successively_working_fval, iters=self.options['maxiter'])

        optimizer_result = OptimizerResult(
            x=pos,
            fval=np.array(cost)
        )

        return optimizer_result

    def is_least_squares(self):
        return False


class NLoptOptimizer(Optimizer):
    """
    Global/Local optimization using NLopt.
    Package homepage: https://nlopt.readthedocs.io/en/latest/
    """

    def __init__(self, method=None, local_method=None, options: Dict = None,
                 local_options: Dict = None):
        """
        Parameters
        ----------
        method:
            Local or global Optimizer to use for minimization.
        local_method:
            Local method to use in combination with the global optimizer (
            for the MLSL family of solvers) or to solve a subproblem (for the
            AUGLAG family of solvers)
        options:
            Optimizer options. scipy option `maxiter` is automatically
            transformed into `maxeval` and takes precedence.
        local_options:
            Optimizer options for the local method
        """

        super().__init__()

        if options is None:
            options = {}
        elif 'maxiter' in options:
            options['maxeval'] = options.pop('maxiter')
        if local_options is None:
            local_options = {}
        self.options = options
        self.local_options = local_options

        if nlopt is None:
            raise ImportError(
                "This optimizer requires an installation of NLopt. You can "
                "install NLopt via `pip install nlopt`.")

        if method is None:
            method = nlopt.LD_LBFGS

        needs_local_method = [
            nlopt.G_MLSL, nlopt.G_MLSL_LDS, nlopt.GD_MLSL, nlopt.GD_MLSL_LDS,
            nlopt.AUGLAG, nlopt.AUGLAG_EQ
        ]

        if local_method is None and method in needs_local_method:
            local_method = nlopt.LD_LBFGS

        if local_method is not None and method not in needs_local_method:
            raise ValueError(f'Method "{method}" does not allow a local '
                             f'method. Please set `local_method` to None.')

        self.local_methods = [
            nlopt.LD_VAR1, nlopt.LD_VAR2, nlopt.LD_TNEWTON_PRECOND_RESTART,
            nlopt.LD_TNEWTON_PRECOND, nlopt.LD_TNEWTON_RESTART,
            nlopt.LD_TNEWTON, nlopt.LD_LBFGS,
            nlopt.LD_SLSQP, nlopt.LD_CCSAQ, nlopt.LD_MMA, nlopt.LN_SBPLX,
            nlopt.LN_NELDERMEAD, nlopt.LN_PRAXIS, nlopt.LN_NEWUOA,
            nlopt.LN_NEWUOA_BOUND, nlopt.LN_BOBYQA, nlopt.LN_COBYLA,
        ]
        self.global_methods = [
            nlopt.GN_ESCH, nlopt.GN_ISRES, nlopt.GN_AGS, nlopt.GD_STOGO,
            nlopt.GD_STOGO_RAND, nlopt.G_MLSL, nlopt.G_MLSL_LDS, nlopt.GD_MLSL,
            nlopt.GD_MLSL_LDS, nlopt.GN_CRS2_LM, nlopt.GN_ORIG_DIRECT,
            nlopt.GN_ORIG_DIRECT_L, nlopt.GN_DIRECT, nlopt.GN_DIRECT_L,
            nlopt.GN_DIRECT_L_NOSCAL, nlopt.GN_DIRECT_L_RAND,
            nlopt.GN_DIRECT_L_RAND_NOSCAL,
        ]
        self.hybrid_methods = [
            nlopt.AUGLAG, nlopt.AUGLAG_EQ
        ]
        methods = self.local_methods + self.global_methods + \
            self.hybrid_methods

        if method not in methods:
            raise ValueError(f'"{method}" is not a valid method. Valid '
                             f'methods are: {methods}')

        self.method = method

        if local_method is not None and local_method not in self.local_methods:
            raise ValueError(f'"{local_method}" is not a valid method. Valid '
                             f'methods are: {self.local_methods}')

        self.local_method = local_method

    @fix_decorator
    @time_decorator
    @history_decorator
    def minimize(
            self,
            problem: Problem,
            x0: np.ndarray,
            id: str,
            history_options: HistoryOptions = None,
    ) -> OptimizerResult:

        opt = nlopt.opt(self.method, problem.dim)

        valid_options = ['ftol_abs', 'ftol_rel', 'xtol_abs', 'xtol_rel',
                         'stopval', 'x_weights', 'maxeval', 'maxtime',
                         'initial_step']

        def set_options(o, options):
            for option, value in options.items():
                if option not in valid_options:
                    raise ValueError(
                        f'"{option}" is not a valid option. Valid '
                        f'options are: {valid_options}')
                getattr(o, f'set_{option}')(value)

        if self.local_method is not None:
            local_opt = nlopt.opt(self.local_method, problem.dim)
            set_options(local_opt, self.local_options)
            opt.set_local_optimizer(local_opt)

        if self.method in self.global_methods:
            check_finite_bounds(problem.ub, problem.lb)

        opt.set_lower_bounds(problem.lb)
        opt.set_upper_bounds(problem.ub)

        def nlopt_objective(x, grad):
            if grad.size > 0:
                sensi_orders = (0, 1)
            else:
                sensi_orders = (0,)
            r = problem.objective(x, sensi_orders, MODE_FUN, True)
            if grad.size > 0:
                grad[:] = r[GRAD]  # note that this must be inplace
            return r[FVAL]

        opt.set_min_objective(nlopt_objective)

        set_options(opt, self.options)
        try:
            result = opt.optimize(x0)
            msg = 'Finished Successfully.'
        except (nlopt.RoundoffLimited, nlopt.ForcedStop, ValueError,
                RuntimeError, MemoryError) as e:
            result = None
            msg = str(e)
        optimizer_result = OptimizerResult(x=result,
                                           fval=opt.last_optimum_value(),
                                           message=msg,
                                           exitflag=opt.last_optimize_result())

        return optimizer_result

    def is_least_squares(self):
        return False


class FidesOptimizer(Optimizer):
    """
    Global/Local optimization using the trust region optimizer fides.
    Package Homepage: https://fides-optimizer.readthedocs.io/en/latest
    """

    def __init__(
            self,
            hessian_update: Optional['HessianApproximation'] = 'Hybrid',
            options: Optional[Dict] = None,
            verbose: Optional[int] = logging.INFO
    ):
        """
        Parameters
        ----------
        options:
            Optimizer options.
        hessian_update:
            Hessian update strategy. If this is None, a hybrid approximation
            that switches from the problem.objective provided Hessian (
            approximation) to a BFGS approximation will be used.
        """

        super().__init__()

        if hessian_update == 'Hybrid':
            hessian_update = fides.HybridFixed()

        if hessian_update is not None and \
                not isinstance(hessian_update, HessianApproximation):
            raise ValueError('Incompatible type for hessian update. '
                             'Must be a HessianApproximation, '
                             f'was {type(hessian_update)}.')

        if options is None:
            options = {}

        self.verbose = verbose
        self.options = options
        self.hessian_update = hessian_update

    @fix_decorator
    @time_decorator
    @history_decorator
    def minimize(
            self,
            problem: Problem,
            x0: np.ndarray,
            id: str,
            history_options: HistoryOptions = None,
    ) -> OptimizerResult:

        if fides is None:
            raise ImportError(
                "This optimizer requires an installation of fides. You can "
                "install fides via `pip install fides`."
            )

        resfun = self.hessian_update.requires_resfun if self.hessian_update \
            is not None else False

        args = {'mode': MODE_RES if resfun else MODE_FUN}

        if not problem.objective.has_grad:
            raise ValueError('Fides cannot be applied to problems '
                             'with objectives that do not support '
                             'gradient evaluation.')

        if self.hessian_update is None or (
            self.hessian_update.requires_hess and not resfun
        ):
            if not problem.objective.has_hess:
                raise ValueError('Specified hessian update scheme cannot be '
                                 'used with objectives that do not support '
                                 'Hessian computation.')
            args['sensi_orders'] = (0, 1, 2)
        else:
            args['sensi_orders'] = (0, 1)

        opt = fides.Optimizer(
            fun=problem.objective, funargs=args, ub=problem.ub, lb=problem.lb,
            verbose=self.verbose, hessian_update=self.hessian_update,
            options=self.options, resfun=resfun
        )

        try:
            opt.minimize(x0)
            if opt.converged:
                msg = 'Finished Successfully.'
            else:
                msg = 'Failed to converge'
        except RuntimeError as err:
            msg = str(err)

        optimizer_result = OptimizerResult(
            x=opt.x_min, fval=opt.fval_min if not resfun else None,
            grad=opt.grad_min, hess=opt.hess, message=msg,
            exitflag=opt.exitflag
        )

        return optimizer_result

    def is_least_squares(self):
        return False



from typing import Dict, Union


class OptimizeOptions(dict):
    """
    Options for the multistart optimization.

    Parameters
    ----------
    startpoint_resample:
        Flag indicating whether initial points are supposed to be resampled if
        function evaluation fails at the initial point
    allow_failed_starts:
        Flag indicating whether we tolerate that exceptions are thrown during
        the minimization process.
    """

    def __init__(self,
                 startpoint_resample: bool = False,
                 allow_failed_starts: bool = True):
        super().__init__()

        self.startpoint_resample: bool = startpoint_resample
        self.allow_failed_starts: bool = allow_failed_starts

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(key)

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    @staticmethod
    def assert_instance(
            maybe_options: Union['OptimizeOptions', Dict]
    ) -> 'OptimizeOptions':
        """
        Returns a valid options object.

        Parameters
        ----------

        maybe_options: OptimizeOptions or dict
        """
        if isinstance(maybe_options, OptimizeOptions):
            return maybe_options
        options = OptimizeOptions(**maybe_options)
        return options



"""
Prediction
==========

Generate predictions from simulations with specified parameter vectors, with
optional post-processing.
"""


from .amici_predictor import AmiciPredictor
from .result import (
    PredictionResult,
    PredictionConditionResult,
)
from .task import PredictorTask



import abc


class Task(abc.ABC):
    """
    A task is one of a list of independent
    execution tasks that are submitted to the execution engine
    to be executed using the execute() method, commonly in parallel.
    """

    def __init__(self):
        pass

    @abc.abstractmethod
    def execute(self):
        """
        Execute the task and return its results.
        """



"""Constants used in the `pypesto.predict` module."""


MODE_FUN = 'mode_fun'  # mode for function values
MODE_RES = 'mode_res'  # mode for residuals

OUTPUT_IDS = 'output_ids'  # data member in PredictionConditionResult
PARAMETER_IDS = 'x_names'  # data member in PredictionConditionResult
CONDITION_IDS = 'condition_ids'
TIMEPOINTS = 'timepoints'  # data member in PredictionConditionResult
OUTPUT = 'output'  # field in the return dict of AmiciPredictor
OUTPUT_SENSI = 'output_sensi'  # field in the return dict of AmiciPredictor
OUTPUT_WEIGHT = 'output_weight'  # field in the return dict of AmiciPredictor
OUTPUT_SIGMAY = 'output_sigmay'  # field in the return dict of AmiciPredictor

# separator in the conditions_ids betweeen preequilibration and simulation
# condition
CONDITION_SEP = '::'

RDATAS = 'rdatas'  # return field of call to pypesto objective
AMICI_T = 't'  # return field in amici simulation result
AMICI_X = 'x'  # return field in amici simulation result
AMICI_SX = 'sx'  # return field in amici simulation result
AMICI_Y = 'y'  # return field in amici simulation result
AMICI_SY = 'sy'  # return field in amici simulation result
AMICI_LLH = 'llh'  # return field in amici simulation result
AMICI_STATUS = 'status'  # return field in amici simulation result
AMICI_SIGMAY = 'sigmay'  # return field in amici simulation result

CSV = 'csv'  # return file format
H5 = 'h5'  # return file format
TIME = 'time'  # column name in returned csv

CONDITION = 'condition'
CONDITION_IDS = 'condition_ids'


def get_condition_label(condition_id: str) -> str:
    """Convert a condition ID to a label.

    Labels for conditions are used at different locations (e.g. ensemble
    prediction code, and visualization code). This method ensures that the same
    condition is labeled identically everywhere.

    Parameters
    ----------
    condition_id:
        The condition ID that will be used to generate a label.

    Returns
    -------
    The condition label.
    """
    return f'condition_{condition_id}'



import numpy as np
import pandas as pd
import numbers
import h5py
import copy
import time
import os
import abc
from typing import Any, Dict, List, Tuple, Sequence, Union

from .constants import (
    MODE_FUN, MODE_RES, FVAL, GRAD, HESS, RES, SRES, CHI2, SCHI2, TIME,
    N_FVAL, N_GRAD, N_HESS, N_RES, N_SRES, X)
from .util import (
    res_to_chi2, res_to_fval, sres_to_schi2, sres_to_fim, schi2_to_grad)

ResultDict = Dict[str, Union[float, np.ndarray]]
MaybeArray = Union[np.ndarray, 'np.nan']


def trace_wrap(f):
    """
    Wrapper around trace getters that transforms input `ix` vectors to a valid
    index list, and reduces for integer `ix` the output to a single value.
    """
    def wrapped_f(
            self, ix: Union[Sequence[int], int, None] = None,
            trim: bool = False
    ) -> Union[Sequence[Union[float, MaybeArray]], Union[float, MaybeArray]]:
        # whether to reduce the output
        reduce = isinstance(ix, numbers.Integral)
        # default: full list
        if ix is None:
            if trim:
                ix = self.get_trimmed_indices()
            else:
                ix = np.arange(0, len(self), dtype=int)
        # turn every input into an index list
        if reduce:
            ix = np.array([ix], dtype=int)
        # obtain the trace
        trace = f(self, ix)
        # reduce the output
        if reduce:
            trace = trace[0]
        return trace
    return wrapped_f


class HistoryOptions(dict):
    """
    Options for the objective that are used in optimization, profiling
    and sampling.

    In addition implements a factory pattern to generate history objects.

    Parameters
    ----------
    trace_record:
        Flag indicating whether to record the trace of function calls.
        The trace_record_* flags only become effective if
        trace_record is True.
    trace_record_grad:
        Flag indicating whether to record the gradient in the trace.
    trace_record_hess:
        Flag indicating whether to record the Hessian in the trace.
    trace_record_res:
        Flag indicating whether to record the residual in
        the trace.
    trace_record_sres:
        Flag indicating whether to record the residual sensitivities in
        the trace.
    trace_record_chi2:
        Flag indicating whether to record the chi2 in the trace.
    trace_record_schi2:
        Flag indicating whether to record the chi2 sensitivities in the
        trace.
    trace_save_iter:
        After how many iterations to store the trace.
    storage_file:
        File to save the history to. Can be any of None, a
        "{filename}.csv", or a "{filename}.hdf5" file. Depending on the values,
        the `create_history` method creates the appropriate object.
        Occurrences of "{id}" in the file name are replaced by the `id`
        upon creation of a history, if applicable.
    """

    def __init__(self,
                 trace_record: bool = False,
                 trace_record_grad: bool = True,
                 trace_record_hess: bool = True,
                 trace_record_res: bool = True,
                 trace_record_sres: bool = True,
                 trace_record_chi2: bool = True,
                 trace_record_schi2: bool = True,
                 trace_save_iter: int = 10,
                 storage_file: str = None):
        super().__init__()
        self.trace_record = trace_record
        self.trace_record_grad = trace_record_grad
        self.trace_record_hess = trace_record_hess
        self.trace_record_res = trace_record_res
        self.trace_record_sres = trace_record_sres
        self.trace_record_chi2 = trace_record_chi2
        self.trace_record_schi2 = trace_record_schi2
        self.trace_save_iter = trace_save_iter
        self.storage_file = storage_file

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(key)

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    @staticmethod
    def assert_instance(
            maybe_options: Union['HistoryOptions', Dict]
    ) -> 'HistoryOptions':
        """
        Returns a valid options object.

        Parameters
        ----------
        maybe_options: HistoryOptions or dict
        """
        if isinstance(maybe_options, HistoryOptions):
            return maybe_options
        options = HistoryOptions(**maybe_options)
        return options

    def create_history(
            self, id: str, x_names: Sequence[str]
    ) -> 'History':
        """Factory method creating a :class:`History` object.

        Parameters
        ----------
        id:
            Identifier for the history.
        x_names:
            Parameter names.
        """
        # create different history types based on the inputs

        if self.storage_file is None:
            if self.trace_record:
                return MemoryHistory(options=self)
            else:
                return History(options=self)

        storage_file = self.storage_file.replace("{id}", id)

        _, type = os.path.splitext(storage_file)

        if type == '.csv':
            return CsvHistory(
                x_names=x_names,
                file=storage_file, options=self)
        elif type in ['.hdf5', '.h5']:
            return Hdf5History(id=id, file=storage_file, options=self)
        else:
            raise ValueError(
                "Currently, only history storage to '.csv' and '.hdf5'"
                "is supported")


class HistoryBase(abc.ABC):
    """Abstract base class for history objects.

    Can be used as a dummy history, but does not implement any history
    functionality.
    """

    def __len__(self):
        """Number of stored entries in the history"""
        raise NotImplementedError()

    def update(
            self,
            x: np.ndarray,
            sensi_orders: Tuple[int, ...],
            mode: str,
            result: ResultDict
    ) -> None:
        """Update history after a function evaluation.

        Parameters
        ----------
        x:
            The parameter vector.
        sensi_orders:
            The sensitivity orders computed.
        mode:
            The objective function mode computed (function value or residuals).
        result:
            The objective function values for parameters `x`, sensitivities
            `sensi_orders` and mode `mode`.
        """

    def finalize(self):
        """Finalize history. Called after a run."""

    @property
    def n_fval(self) -> int:
        """Number of function evaluations."""
        raise NotImplementedError()

    @property
    def n_grad(self) -> int:
        """Number of gradient evaluations."""
        raise NotImplementedError()

    @property
    def n_hess(self) -> int:
        """Number of Hessian evaluations."""
        raise NotImplementedError()

    @property
    def n_res(self) -> int:
        """Number of residual evaluations."""
        raise NotImplementedError()

    @property
    def n_sres(self) -> int:
        """Number or residual sensitivity evaluations."""
        raise NotImplementedError()

    @property
    def start_time(self) -> float:
        """Start time."""
        raise NotImplementedError()

    def get_x_trace(
            self, ix: Union[int, Sequence[int], None] = None,
            trim: bool = False
    ) -> Union[Sequence[np.ndarray], np.ndarray]:
        """Parameters.

        Takes as parameter an index or indices and returns corresponding trace
        values. If only a single value is requested, the list is flattened.
        """
        raise NotImplementedError()

    def get_fval_trace(
            self, ix: Union[int, Sequence[int], None] = None,
            trim: bool = False
    ) -> Union[Sequence[float], float]:
        """Function values.

        Takes as parameter an index or indices and returns corresponding trace
        values. If only a single value is requested, the list is flattened.
        """
        raise NotImplementedError()

    def get_grad_trace(
            self, ix: Union[int, Sequence[int], None] = None,
            trim: bool = False
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        """Gradients.

        Takes as parameter an index or indices and returns corresponding trace
        values. If only a single value is requested, the list is flattened.
        """
        raise NotImplementedError()

    def get_hess_trace(
            self, ix: Union[int, Sequence[int], None] = None,
            trim: bool = False
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        """Hessians.

        Takes as parameter an index or indices and returns corresponding trace
        values. If only a single value is requested, the list is flattened.
        """
        raise NotImplementedError()

    def get_res_trace(
            self, ix: Union[int, Sequence[int], None] = None,
            trim: bool = False
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        """Residuals.

        Takes as parameter an index or indices and returns corresponding trace
        values. If only a single value is requested, the list is flattened.
        """
        raise NotImplementedError()

    def get_sres_trace(
            self, ix: Union[int, Sequence[int], None] = None,
            trim: bool = False
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        """Residual sensitivities.

        Takes as parameter an index or indices and returns corresponding trace
        values. If only a single value is requested, the list is flattened.
        """
        raise NotImplementedError()

    def get_chi2_trace(
            self, ix: Union[int, Sequence[int], None] = None,
            trim: bool = False
    ) -> Union[Sequence[float], float]:
        """Chi2 values.

        Takes as parameter an index or indices and returns corresponding trace
        values. If only a single value is requested, the list is flattened.
        """
        raise NotImplementedError()

    def get_schi2_trace(
            self, ix: Union[int, Sequence[int], None] = None,
            trim: bool = False
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        """Chi2 sensitivities.

        Takes as parameter an index or indices and returns corresponding trace
        values. If only a single value is requested, the list is flattened.
        """
        raise NotImplementedError()

    def get_time_trace(
            self, ix: Union[int, Sequence[int], None] = None,
            trim: bool = False
    ) -> Union[Sequence[float], float]:
        """Cumulative execution times.

        Takes as parameter an index or indices and returns corresponding trace
        values. If only a single value is requested, the list is flattened.
        """
        raise NotImplementedError()

    def get_trimmed_indices(self):
        """
        Returns indices of the history to get a monotonically
        decreasing history.
        """
        fval_trace = self.get_fval_trace()
        return np.where(fval_trace <= np.fmin.accumulate(fval_trace))[0]


class History(HistoryBase):
    """Tracks numbers of function evaluations only, no trace.

    Parameters
    ----------
    options:
        History options.
    """

    def __init__(self, options: Union[HistoryOptions, Dict] = None):
        self._n_fval: int = 0
        self._n_grad: int = 0
        self._n_hess: int = 0
        self._n_res: int = 0
        self._n_sres: int = 0
        self._start_time = time.time()

        if options is None:
            options = HistoryOptions()
        options = HistoryOptions.assert_instance(options)
        self.options: HistoryOptions = options

    def update(
            self,
            x: np.ndarray,
            sensi_orders: Tuple[int, ...],
            mode: str,
            result: ResultDict
    ) -> None:
        """Update history after a function evaluation.

        Parameters
        ----------
        x:
            The parameter vector.
        sensi_orders:
            The sensitivity orders computed.
        mode:
            The objective function mode computed (function value or residuals).
        result:
            The objective function values for parameters `x`, sensitivities
            `sensi_orders` and mode `mode`.
        """
        res = result.get(RES, None)
        if res is not None and FVAL not in result:
            # no option trace_record_fval
            result[FVAL] = res_to_fval(res)
        self._update_counts(sensi_orders, mode)

    def finalize(self):
        pass

    def _update_counts(self,
                       sensi_orders: Tuple[int, ...],
                       mode: str):
        """
        Update the counters.
        """
        if mode == MODE_FUN:
            if 0 in sensi_orders:
                self._n_fval += 1
            if 1 in sensi_orders:
                self._n_grad += 1
            if 2 in sensi_orders:
                self._n_hess += 1
        elif mode == MODE_RES:
            if 0 in sensi_orders:
                self._n_res += 1
            if 1 in sensi_orders:
                self._n_sres += 1

    @property
    def n_fval(self) -> int:
        return self._n_fval

    @property
    def n_grad(self) -> int:
        return self._n_grad

    @property
    def n_hess(self) -> int:
        return self._n_hess

    @property
    def n_res(self) -> int:
        return self._n_res

    @property
    def n_sres(self) -> int:
        return self._n_sres

    @property
    def start_time(self) -> float:
        return self._start_time


class MemoryHistory(History):
    """Tracks numbers of function evaluations and keeps an in-memory
    trace of function evaluations.

    Parameters
    ----------
    options:
        History options.
    """

    def __init__(self, options: Union[HistoryOptions, Dict] = None):
        super().__init__(options=options)
        self._trace_keys = {X, FVAL, GRAD, HESS, RES, SRES, CHI2, SCHI2, TIME}
        self._trace: Dict[str, Any] = {key: [] for key in self._trace_keys}

    def __len__(self):
        return len(self._trace[TIME])

    def update(
            self,
            x: np.ndarray,
            sensi_orders: Tuple[int, ...],
            mode: str,
            result: ResultDict
    ) -> None:
        super().update(x, sensi_orders, mode, result)
        self._update_trace(x, mode, result)

    def _update_trace(self, x, mode, result):
        """Update internal trace representation."""
        ret = extract_values(mode, result, self.options)
        for key in self._trace_keys - {X, TIME}:
            self._trace[key].append(ret[key])
        used_time = time.time() - self._start_time
        self._trace[X].append(x)
        self._trace[TIME].append(used_time)

    @trace_wrap
    def get_x_trace(
            self, ix: Union[int, Sequence[int], None] = None,
            trim: bool = False
    ) -> Union[Sequence[np.ndarray], np.ndarray]:
        return [self._trace[X][i] for i in ix]

    @trace_wrap
    def get_fval_trace(
            self, ix: Union[int, Sequence[int], None] = None,
            trim: bool = False
    ) -> Union[Sequence[float], float]:
        return [self._trace[FVAL][i] for i in ix]

    @trace_wrap
    def get_grad_trace(
            self, ix: Union[int, Sequence[int], None] = None,
            trim: bool = False
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        return [self._trace[GRAD][i] for i in ix]

    @trace_wrap
    def get_hess_trace(
            self, ix: Union[int, Sequence[int], None] = None,
            trim: bool = False
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        return [self._trace[HESS][i] for i in ix]

    @trace_wrap
    def get_res_trace(
            self, ix: Union[int, Sequence[int], None] = None,
            trim: bool = False
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        return [self._trace[RES][i] for i in ix]

    @trace_wrap
    def get_sres_trace(
            self, ix: Union[int, Sequence[int], None] = None,
            trim: bool = False
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        return [self._trace[SRES][i] for i in ix]

    @trace_wrap
    def get_chi2_trace(
            self, ix: Union[int, Sequence[int], None] = None,
            trim: bool = False
    ) -> Union[Sequence[float], float]:
        return [self._trace[CHI2][i] for i in ix]

    @trace_wrap
    def get_schi2_trace(
            self, ix: Union[int, Sequence[int], None] = None,
            trim: bool = False
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        return [self._trace[SCHI2][i] for i in ix]

    @trace_wrap
    def get_time_trace(
            self, ix: Union[int, Sequence[int], None] = None,
            trim: bool = False
    ) -> Union[Sequence[float], float]:
        return [self._trace[TIME][i] for i in ix]


class CsvHistory(History):
    """Stores a representation of the history in a CSV file.

    Parameters
    ----------
    file:
        CSV file name.
    x_names:
        Parameter names.
    options:
        History options.
    load_from_file:
        If True, history will be initialized from data in the specified file
    """

    def __init__(self,
                 file: str,
                 x_names: Sequence[str] = None,
                 options: Union[HistoryOptions, Dict] = None,
                 load_from_file: bool = False):
        super().__init__(options=options)
        self.x_names = x_names
        self._trace: Union[pd.DataFrame, None] = None
        self.file = os.path.abspath(file)

        # create trace file dirs
        if self.file is not None:
            dirname = os.path.dirname(self.file)
            os.makedirs(dirname, exist_ok=True)

        if load_from_file:
            trace = pd.read_csv(self.file, header=[0, 1], index_col=0)
            # replace 'nan' in cols with np.NAN
            cols = pd.DataFrame(trace.columns.to_list())
            cols[cols == 'nan'] = np.NaN
            trace.columns = pd.MultiIndex.from_tuples(
                cols.to_records(index=False).tolist()
            )
            for col in trace.columns:
                # transform strings to np.ndarrays
                trace[col] = trace[col].apply(string2ndarray)

            self._trace = trace
            self.x_names = trace[X].columns
            self._update_counts_from_trace()

    def __len__(self):
        return len(self._trace)

    def _update_counts_from_trace(self):
        self._n_fval = self._trace[('n_fval', np.NaN)].max()
        self._n_grad = self._trace[('n_grad', np.NaN)].max()
        self._n_hess = self._trace[('n_hess', np.NaN)].max()
        self._n_res = self._trace[('n_res', np.NaN)].max()
        self._n_sres = self._trace[('n_sres', np.NaN)].max()

    def update(
            self,
            x: np.ndarray,
            sensi_orders: Tuple[int, ...],
            mode: str,
            result: ResultDict
    ) -> None:
        super().update(x, sensi_orders, mode, result)
        self._update_trace(x, mode, result)

    def finalize(self):
        """Finalize history. Called after a run."""
        super().finalize()
        self._save_trace(finalize=True)

    def _update_trace(self,
                      x: np.ndarray,
                      mode: str,
                      result: ResultDict):
        """
        Update and possibly store the trace.
        """
        if not self.options.trace_record:
            return

        # init trace
        if self._trace is None:
            self._init_trace(x)

        # extract function values
        ret = extract_values(mode, result, self.options)

        used_time = time.time() - self._start_time

        # create table row
        row = pd.Series(name=len(self._trace),
                        index=self._trace.columns,
                        dtype='object')

        values = {
            TIME: used_time,
            N_FVAL: self._n_fval,
            N_GRAD: self._n_grad,
            N_HESS: self._n_hess,
            N_RES: self._n_res,
            N_SRES: self._n_sres,
            FVAL: ret[FVAL],
            RES: ret[RES],
            SRES: ret[SRES],
            CHI2: ret[CHI2],
            HESS: ret[HESS],
        }

        for var, val in values.items():
            row[(var, float('nan'))] = val

        for var, val in {X: x, GRAD: ret[GRAD], SCHI2: ret[SCHI2]}.items():
            if var == X or self.options[f'trace_record_{var}']:
                row[var] = val
            else:
                row[(var, float('nan'))] = np.NaN

        self._trace = self._trace.append(row)

        # save trace to file
        self._save_trace()

    def _init_trace(self, x: np.ndarray):
        """
        Initialize the trace.
        """
        if self.x_names is None:
            self.x_names = [f'x{i}' for i, _ in enumerate(x)]

        columns: List[Tuple] = [
            (c, float('nan')) for c in [
                TIME, N_FVAL, N_GRAD, N_HESS, N_RES, N_SRES,
                FVAL, CHI2, RES, SRES, HESS,
            ]
        ]

        for var in [X, GRAD, SCHI2]:
            if var == 'x' or self.options[f'trace_record_{var}']:
                columns.extend([
                    (var, x_name)
                    for x_name in self.x_names
                ])
            else:
                columns.extend([(var,)])

        # TODO: multi-index for res, sres, hess
        self._trace = pd.DataFrame(columns=pd.MultiIndex.from_tuples(columns),
                                   dtype='float64')

        # only non-float64
        trace_dtypes = {
            RES: 'object',
            SRES: 'object',
            HESS: 'object',
            N_FVAL: 'int64',
            N_GRAD: 'int64',
            N_HESS: 'int64',
            N_RES: 'int64',
            N_SRES: 'int64',
        }

        for var, dtype in trace_dtypes.items():
            self._trace[(var, np.NaN)] = \
                self._trace[(var, np.NaN)].astype(dtype)

    def _save_trace(self, finalize: bool = False):
        """
        Save to file via pd.DataFrame.to_csv() if `self.storage_file` is
        not None and other conditions apply.
        """
        if self.file is None:
            return

        if finalize \
                or (len(self._trace) > 0 and len(self._trace) %
                    self.options.trace_save_iter == 0):
            # save
            trace_copy = copy.deepcopy(self._trace)
            for field in [('hess', np.NaN), ('res', np.NaN), ('sres', np.NaN)]:
                trace_copy[field] = trace_copy[field].apply(
                    ndarray2string_full
                )
            trace_copy.to_csv(self.file)

    @trace_wrap
    def get_x_trace(
            self, ix: Union[int, Sequence[int], None] = None,
            trim: bool = False
    ) -> Union[Sequence[np.ndarray], np.ndarray]:
        return list(self._trace[X].values[ix])

    @trace_wrap
    def get_fval_trace(
            self, ix: Union[int, Sequence[int], None],
            trim: bool = False
    ) -> Union[Sequence[float], float]:
        return list(self._trace[(FVAL, np.nan)].values[ix])

    @trace_wrap
    def get_grad_trace(
            self, ix: Union[int, Sequence[int], None] = None,
            trim: bool = False
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        return list(self._trace[GRAD].values[ix])

    @trace_wrap
    def get_hess_trace(
            self, ix: Union[int, Sequence[int], None] = None,
            trim: bool = False
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        return list(self._trace[(HESS, np.nan)].values[ix])

    @trace_wrap
    def get_res_trace(
            self, ix: Union[int, Sequence[int], None] = None,
            trim: bool = False
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        return list(self._trace[(RES, np.nan)].values[ix])

    @trace_wrap
    def get_sres_trace(
            self, ix: Union[int, Sequence[int], None] = None,
            trim: bool = False
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        return list(self._trace[(SRES, np.nan)].values[ix])

    @trace_wrap
    def get_chi2_trace(
            self, ix: Union[int, Sequence[int], None] = None,
            trim: bool = False
    ) -> Union[Sequence[float], float]:
        return list(self._trace[(CHI2, np.nan)].values[ix])

    @trace_wrap
    def get_schi2_trace(
            self, ix: Union[int, Sequence[int], None] = None,
            trim: bool = False
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        return list(self._trace[SCHI2].values[ix])

    @trace_wrap
    def get_time_trace(
            self, ix: Union[int, Sequence[int], None] = None,
            trim: bool = False
    ) -> Union[Sequence[float], float]:
        return list(self._trace[(TIME, np.nan)].values[ix])


class Hdf5History(History):
    """Stores a representation of the history in an HDF5 file.

    Parameters
    ----------
    id:
        Id of the history
    file:
        HDF5 file name.
    options:
        History options.
    """

    def __init__(self,
                 id: str,
                 file: str,
                 options: Union[HistoryOptions, Dict] = None):
        super().__init__(options=options)
        self.id = id
        self.file = file
        self._generate_hdf5_group()

    def __len__(self):
        with h5py.File(self.file, 'r') as f:
            return f[f'history/{self.id}/trace/'].attrs[
                        'n_iterations']

    def update(
            self,
            x: np.ndarray,
            sensi_orders: Tuple[int, ...],
            mode: str,
            result: ResultDict
    ) -> None:
        super().update(x, sensi_orders, mode, result)
        self._update_trace(x, sensi_orders, mode, result)

    def get_history_directory(self):
        return self.file

    def finalize(self):
        super().finalize()

    @staticmethod
    def load(id: str,
             file: str):
        """Loads the History object from memory."""
        loaded_h5history = Hdf5History(id, file)
        loaded_h5history._recover_options(file)
        return loaded_h5history

    def _recover_options(self, file: str):
        """
        Recovers options when loading the hdf5 history from memory
        by testing which entries were recorded.
        """
        trace_record = self._check_for_not_nan_entries(X)
        trace_record_grad = self._check_for_not_nan_entries(GRAD)
        trace_record_hess = self._check_for_not_nan_entries(HESS)
        trace_record_res = self._check_for_not_nan_entries(RES)
        trace_record_sres = self._check_for_not_nan_entries(SRES)
        trace_record_chi2 = self._check_for_not_nan_entries(CHI2)
        trace_record_schi2 = self._check_for_not_nan_entries(SCHI2)
        storage_file = file

        restored_history_options = \
            HistoryOptions(trace_record=trace_record,
                           trace_record_grad=trace_record_grad,
                           trace_record_hess=trace_record_hess,
                           trace_record_res=trace_record_res,
                           trace_record_sres=trace_record_sres,
                           trace_record_chi2=trace_record_chi2,
                           trace_record_schi2=trace_record_schi2,
                           trace_save_iter=self.trace_save_iter,
                           storage_file=storage_file)

        self.options = restored_history_options

    def _check_for_not_nan_entries(self, hdf5_group: str) -> bool:
        """Checks if there exist not-nan entries stored for a given group"""
        group = self._get_hdf5_entries(hdf5_group, ix=None)

        for entry in group:
            if not (entry is None or np.all(np.isnan(entry))):
                return True

        return False

    # overwrite _update_counts
    def _update_counts(self,
                       sensi_orders: Tuple[int, ...],
                       mode: str):
        """
        Update the counters in the hdf5
        """
        with h5py.File(self.file, 'a') as f:

            if mode == MODE_FUN:
                if 0 in sensi_orders:
                    f[f'history/{self.id}/trace/'].attrs[
                        'n_fval'] += 1
                if 1 in sensi_orders:
                    f[f'history/{self.id}/trace/'].attrs[
                        'n_grad'] += 1
                if 2 in sensi_orders:
                    f[f'history/{self.id}/trace/'].attrs[
                        'n_hess'] += 1
            elif mode == MODE_RES:
                if 0 in sensi_orders:
                    f[f'history/{self.id}/trace/'].attrs[
                        'n_res'] += 1
                if 1 in sensi_orders:
                    f[f'history/{self.id}/trace/'].attrs[
                        'n_sres'] += 1

    @property
    def n_fval(self) -> int:
        with h5py.File(self.file, 'r') as f:
            return f[f'history/{self.id}/trace/'].attrs['n_fval']

    @property
    def n_grad(self) -> int:
        with h5py.File(self.file, 'r') as f:
            return f[f'history/{self.id}/trace/'].attrs['n_grad']

    @property
    def n_hess(self) -> int:
        with h5py.File(self.file, 'r') as f:
            return f[f'history/{self.id}/trace/'].attrs['n_hess']

    @property
    def n_res(self) -> int:
        with h5py.File(self.file, 'r') as f:
            return f[f'history/{self.id}/trace/'].attrs['n_res']

    @property
    def n_sres(self) -> int:
        with h5py.File(self.file, 'r') as f:
            return f[f'history/{self.id}/trace/'].attrs['n_sres']

    @property
    def trace_save_iter(self):
        with h5py.File(self.file, 'r') as f:
            return f[f'history/{self.id}/trace/']\
                .attrs['trace_save_iter']

    def _update_trace(self,
                      x: np.ndarray,
                      sensi_orders: Tuple[int],
                      mode: str,
                      result: ResultDict):
        """
        Update and possibly store the trace.
        """

        if not self.options.trace_record:
            return

        # extract function values
        ret = extract_values(mode, result, self.options)

        used_time = time.time() - self._start_time

        values = {
            TIME: used_time,
            X: x,
            FVAL: ret[FVAL],
            GRAD: ret[GRAD],
            RES: ret[RES],
            SRES: ret[SRES],
            CHI2: ret[CHI2],
            SCHI2: ret[SCHI2],
            HESS: ret[HESS],
        }

        with h5py.File(self.file, 'a') as f:

            iteration = f[f'history/{self.id}/trace/'].attrs[
                'n_iterations']

            for key in values.keys():
                if values[key] is not None:
                    f[f'history/{self.id}/trace/'
                      f'{str(iteration)}/{key}'] = values[key]

            f[f'history/{self.id}/trace/'].attrs[
                'n_iterations'] += 1

    def _generate_hdf5_group(self, f: h5py.File = None):
        """
        Generates the group in the hdf5 file, if it does not exist yet.
        """
        try:
            with h5py.File(self.file, 'a') as f:
                if f'history/{self.id}/trace/' not in f:
                    grp = f.create_group(f'history/{self.id}/trace/')
                    grp.attrs['n_iterations'] = 0
                    grp.attrs['n_fval'] = 0
                    grp.attrs['n_grad'] = 0
                    grp.attrs['n_hess'] = 0
                    grp.attrs['n_res'] = 0
                    grp.attrs['n_sres'] = 0
                    grp.attrs['trace_save_iter'] = self.options.trace_save_iter
        except OSError:
            pass

    def _get_hdf5_entries(
        self,
        entry_id: str,
        ix: Union[int, Sequence[int], None] = None,
    ) -> Sequence:
        """Get entries for field `entry_id` from HDF5 file, for indices `ix`.

        Parameters
        ------------
        entry_id:
            The key whose trace is returned.
        ix:
            Index or list of indices of the iterations that will produce
            the trace.
        Returns
        --------
        The entries ix for the key entry_id.
        """
        if ix is None:
            ix = range(len(self))
        trace_result = []

        with h5py.File(self.file, 'r') as f:

            for iteration in ix:
                try:
                    entry = np.array(f[f'history/{self.id}/trace'
                                       f'/{str(iteration)}/{entry_id}'])
                    trace_result.append(entry)
                except KeyError:
                    trace_result.append(None)

        return trace_result

    @trace_wrap
    def get_x_trace(
            self, ix: Union[int, Sequence[int], None] = None,
            trim: bool = False
    ) -> Union[Sequence[np.ndarray], np.ndarray]:
        return self._get_hdf5_entries(X, ix)

    @trace_wrap
    def get_fval_trace(
            self, ix: Union[int, Sequence[int], None] = None,
            trim: bool = False
    ) -> Union[Sequence[float], float]:
        return self._get_hdf5_entries(FVAL, ix)

    @trace_wrap
    def get_grad_trace(
            self, ix: Union[int, Sequence[int], None] = None,
            trim: bool = False
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        return self._get_hdf5_entries(GRAD, ix)

    @trace_wrap
    def get_hess_trace(
            self, ix: Union[int, Sequence[int], None] = None,
            trim: bool = False
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        return self._get_hdf5_entries(HESS, ix)

    @trace_wrap
    def get_res_trace(
            self, ix: Union[int, Sequence[int], None] = None,
            trim: bool = False
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        return self._get_hdf5_entries(RES, ix)

    @trace_wrap
    def get_sres_trace(
            self, ix: Union[int, Sequence[int], None] = None,
            trim: bool = False
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        return self._get_hdf5_entries(SRES, ix)

    @trace_wrap
    def get_chi2_trace(
            self, ix: Union[int, Sequence[int], None] = None,
            trim: bool = False
    ) -> Union[Sequence[float], float]:
        return self._get_hdf5_entries(CHI2, ix)

    @trace_wrap
    def get_schi2_trace(
            self, ix: Union[int, Sequence[int], None] = None,
            trim: bool = False
    ) -> Union[Sequence[MaybeArray], MaybeArray]:
        return self._get_hdf5_entries(SCHI2, ix)

    @trace_wrap
    def get_time_trace(
            self, ix: Union[int, Sequence[int], None] = None,
            trim: bool = False
    ) -> Union[Sequence[float], float]:
        return self._get_hdf5_entries(TIME, ix)


class OptimizerHistory:
    """
    Objective call history. Container around a History object, which keeps
    track of optimal values.

    Attributes
    ----------
    fval0, fval_min:
        Initial and best function value found.
    chi20, chi2_min:
        Initial and best chi2 value found.
    x0, x_min:
        Initial and best parameters found.
    grad_min:
        gradient for best parameters
    hess_min:
        hessian (approximation) for best parameters
    res_min:
        residuals for best parameters
    sres_min:
        residual sensitivities for best parameters

    Parameters
    ----------
    history:
        History object to attach to this container. This history object
        implements the storage of the actual history.
    x0:
        Initial values for optimization
    generate_from_history:
        If set to true, this function will try to fill attributes of this
        function based on the provided history
    """

    def __init__(self,
                 history: History,
                 x0: np.ndarray,
                 generate_from_history: bool = False) -> None:
        self.history: History = history

        # initial point
        self.fval0: Union[float, None] = None
        self.x0: np.ndarray = x0

        # minimum point
        self.fval_min: float = np.inf
        self.x_min: Union[np.ndarray, None] = None
        self.grad_min: Union[np.ndarray, None] = None
        self.hess_min: Union[np.ndarray, None] = None
        self.res_min: Union[np.ndarray, None] = None
        self.sres_min: Union[np.ndarray, None] = None

        if generate_from_history:
            self._compute_vals_from_trace()

    def update(self,
               x: np.ndarray,
               sensi_orders: Tuple[int],
               mode: str,
               result: ResultDict) -> None:
        """Update history and best found value."""
        self.history.update(x, sensi_orders, mode, result)
        self._update_vals(x, result)

    def finalize(self):
        self.history.finalize()

    def _update_vals(self,
                     x: np.ndarray,
                     result: ResultDict):
        """
        Update initial and best function values.
        """
        # update initial point
        if np.allclose(x, self.x0):
            if self.fval0 is None:
                self.fval0 = result.get(FVAL, None)
            self.x0 = x

        # update best point
        fval = result.get(FVAL, None)
        grad = result.get(GRAD, None)
        hess = result.get(HESS, None)
        res = result.get(RES, None)
        sres = result.get(SRES, None)

        if fval is not None and fval < self.fval_min:
            self.fval_min = fval
            self.x_min = x
            self.grad_min = grad
            self.hess_min = hess
            self.res_min = res
            self.sres_min = sres

        # sometimes sensitivities are evaluated on subsequent calls. We can
        # identify this situation by checking that x hasn't changed
        if np.all(self.x_min == x):
            if self.grad_min is None and grad is not None:
                self.grad_min = grad
            if self.hess_min is None and hess is not None:
                self.hess_min = hess
            if self.res_min is None and res is not None:
                self.res_min = res
            if self.sres_min is None and sres is not None:
                self.sres_min = sres

    def _compute_vals_from_trace(self):
        # some optimizers may evaluate hess+grad first to compute trust region
        # etc
        max_init_iter = 3
        for it in range(min(len(self.history), max_init_iter)):
            candidate = self.history.get_fval_trace(it)
            if not np.isnan(candidate) \
                    and np.allclose(self.history.get_x_trace(it), self.x0):
                self.fval0 = float(candidate)
                break

        # we prioritize fval over chi2 as fval is written whenever possible
        ix_min = np.nanargmin(self.history.get_fval_trace())
        # np.argmin returns ndarray when multiple minimal values are found, we
        # generally want the first occurence
        if isinstance(ix_min, np.ndarray):
            ix_min = ix_min[0]

        for var in ['fval', 'chi2', 'x']:
            self.extract_from_history(var, ix_min)
            if var == 'fval':
                self.fval_min = float(self.fval_min)

        if self.history.options.trace_record_res:
            self.extract_from_history('res', ix_min)

        for var in ['grad', 'sres', 'hess']:
            target = f'{var}_min'  # attribute in self we want to set
            ix_try = ix_min + 1  # index we try after ix_min doesnt work
            if getattr(self.history.options, f'trace_record_{var}'):
                self.extract_from_history(var, ix_min)
                if getattr(self, target) is None \
                        and ix_try < len(self.history) \
                        and np.allclose(self.history.get_x_trace(ix_min),
                                        self.history.get_x_trace(ix_try)):
                    # gradient/sres typically evaluated on the next call
                    # so we check if x remains the same and if yes try to
                    # extract from the next
                    self.extract_from_history(var, ix_try)

    def extract_from_history(self, var, ix):
        val = getattr(self.history, f'get_{var}_trace')(ix)
        if not np.all(np.isnan(val)):
            setattr(self, f'{var}_min', val)


def ndarray2string_full(x: Union[np.ndarray, None]) -> Union[str, None]:
    """
    Helper function that converts numpy arrays to string with 16 digit
    numerical precision and no truncation for large arrays

    Parameters
    ----------
    x:
        array to convert

    Returns
    -------
    x:
        array as string
    """
    if not isinstance(x, np.ndarray):
        return x
    return np.array2string(x, threshold=x.size, precision=16,
                           max_line_width=np.inf)


def string2ndarray(x: Union[str, float]) -> Union[np.ndarray, float]:
    """
    Helper function that converts string to numpy arrays

    Parameters
    ----------
    x:
        array to convert

    Returns
    -------
    x:
        array as np.ndarray
    """
    if not isinstance(x, str):
        return x
    if x.startswith('[['):
        return np.vstack([
            np.fromstring(xx, sep=' ')
            for xx in x[2:-2].split(']\n [')
        ])
    else:
        return np.fromstring(x[1:-1], sep=' ')


def extract_values(mode: str,
                   result: ResultDict,
                   options: HistoryOptions) -> Dict:
    """Extract values to record from result."""
    ret = {}
    ret_vars = [FVAL, GRAD, HESS, RES, SRES, CHI2, SCHI2]
    for var in ret_vars:
        if options.get(f'trace_record_{var}', True) and var in result:
            ret[var] = result[var]

    # write values that weren't set yet with alternative methods
    if mode == MODE_RES:
        res_result = result.get(RES, None)
        sres_result = result.get(SRES, None)
        chi2 = res_to_chi2(res_result)
        schi2 = sres_to_schi2(res_result, sres_result)
        fim = sres_to_fim(sres_result)
        alt_values = {CHI2: chi2, SCHI2: schi2, HESS: fim}
        if schi2 is not None:
            alt_values[GRAD] = schi2_to_grad(schi2)

        # filter according to options
        alt_values = {
            key: val
            for key, val in alt_values.items()
            if options.get(f'trace_record_{key}', True)
        }
        for var, val in alt_values.items():
            if val is not None:
                ret[var] = ret.get(var, val)

    # set everything missing to NaN
    for var in ret_vars:
        if var not in ret:
            ret[var] = np.NaN

    return ret



from multiprocessing import Pool
import cloudpickle as pickle
import os
import logging
from typing import List
from tqdm import tqdm

from .base import Engine
from .task import Task


logger = logging.getLogger(__name__)


def work(pickled_task):
    """Unpickle and execute task."""
    task = pickle.loads(pickled_task)
    return task.execute()


class MultiProcessEngine(Engine):
    """
    Parallelize the task execution using multiprocessing.

    Parameters
    ----------
    n_procs:
        The maximum number of processes to use in parallel.
        Defaults to the number of CPUs available on the system according to
        `os.cpu_count()`.
        The effectively used number of processes will be the minimum of
        `n_procs` and the number of tasks submitted.
    """

    def __init__(self, n_procs: int = None):
        super().__init__()

        if n_procs is None:
            n_procs = os.cpu_count()
            logger.warning(
                f"Engine set up to use up to {n_procs} processes in total. "
                f"The number was automatically determined and might not be "
                f"appropriate on some systems.")
        self.n_procs: int = n_procs

    def execute(self, tasks: List[Task], progress_bar: bool = True):
        """Pickle tasks and distribute work over parallel processes.

        Parameters
        ----------
        tasks:
            List of tasks to execute.
        progress_bar:
            Whether to display a progress bar.
        """
        n_tasks = len(tasks)

        pickled_tasks = [pickle.dumps(task) for task in tasks]

        n_procs = min(self.n_procs, n_tasks)
        logger.info(f"Performing parallel task execution on {n_procs} "
                    f"processes.")

        with Pool(processes=n_procs) as pool:
            results = pool.map(work,
                               tqdm(pickled_tasks,
                                    disable=not progress_bar)
                               )

        return results



"""
Sample
======

Draw samples from the distribution, with support for various samplers.
"""

from .sample import sample
from .sampler import (
    InternalSampler,
    Sampler,
)
from .metropolis import (
    MetropolisSampler,
)
from .adaptive_metropolis import (
    AdaptiveMetropolisSampler,
)
from .parallel_tempering import (
    ParallelTemperingSampler,
)
from .adaptive_parallel_tempering import (
    AdaptiveParallelTemperingSampler,
)
from .pymc3 import Pymc3Sampler
from .emcee import EmceeSampler
from .result import McmcPtResult
from .diagnostics import (
    auto_correlation,
    effective_sample_size,
    geweke_test,
)
from .util import (
    calculate_ci_mcmc_sample,
    calculate_ci_mcmc_sample_prediction,
)



import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
import numpy as np

from typing import List, Optional, Tuple, Union, Iterable

from ..result import Result
from ..objective import History
from .reference_points import create_references, ReferencePoint
from .clust_color import assign_colors
from .misc import process_result_list
from .misc import process_y_limits
from .misc import process_offset_y


def optimizer_history(results,
                      ax=None,
                      size=(18.5, 10.5),
                      trace_x='steps',
                      trace_y='fval',
                      scale_y='log10',
                      offset_y=None,
                      colors=None,
                      y_limits=None,
                      start_indices=None,
                      reference=None,
                      legends=None):
    """
    Plot history of optimizer. Can plot either the history of the cost
    function or of the gradient norm, over either the optimizer steps or
    the computation time.

    Parameters
    ----------

    results: pypesto.Result or list
        Optimization result obtained by 'optimize.py' or list of those

    ax: matplotlib.Axes, optional
        Axes object to use.

    size: tuple, optional
        Figure size (width, height) in inches. Is only applied when no ax
        object is specified

    trace_x: str, optional
        What should be plotted on the x-axis?
        Possibilities: 'time', 'steps'
        Default: 'steps'

    trace_y: str, optional
        What should be plotted on the y-axis?
        Possibilities: 'fval', 'gradnorm', 'stepsize'
        Default: 'fval'

    scale_y: str, optional
        May be logarithmic or linear ('log10' or 'lin')

    offset_y: float, optional
        Offset for the y-axis-values, as these are plotted on a log10-scale
        Will be computed automatically if necessary

    colors: list, or RGBA, optional
        list of colors, or single color
        color or list of colors for plotting. If not set, clustering is done
        and colors are assigned automatically

    y_limits: float or ndarray, optional
        maximum value to be plotted on the y-axis, or y-limits

    start_indices: list or int
        list of integers specifying the multistart to be plotted or
        int specifying up to which start index should be plotted

    reference: list, optional
        List of reference points for optimization results, containing et
        least a function value fval

    legends: list or str
        Labels for line plots, one label per result object

    Returns
    -------

    ax: matplotlib.Axes
        The plot axes.
    """
    if isinstance(start_indices, int):
        start_indices = list(range(start_indices))

    # parse input
    (results, colors, legends) = process_result_list(results, colors, legends)

    for j, result in enumerate(results):
        # extract cost function values from result
        (x_label, y_label, vals) = get_trace(result, trace_x, trace_y)

        # compute the necessary offset for the y-axis
        (vals, offset_y, y_label) = get_vals(vals, scale_y, offset_y, y_label,
                                             start_indices)

        # call lowlevel plot routine
        ax = optimizer_history_lowlevel(vals, scale_y=scale_y, ax=ax,
                                        colors=colors[j],
                                        size=size, x_label=x_label,
                                        y_label=y_label,
                                        legend_text=legends[j])

    # parse and apply plotting options
    ref = create_references(references=reference)

    # handle options
    ax = handle_options(ax, vals, ref, y_limits, offset_y)

    return ax


def optimizer_history_lowlevel(vals, scale_y='log10', colors=None, ax=None,
                               size=(18.5, 10.5), x_label='Optimizer steps',
                               y_label='Objective value', legend_text=None):
    """
    Plot optimizer history using list of numpy arrays.

    Parameters
    ----------

    vals: list of numpy arrays
        list of 2xn-arrays (x_values and y_values of the trace)

    scale_y: str, optional
        May be logarithmic or linear ('log10' or 'lin')

    colors: list, or RGBA, optional
        list of colors, or single color
        color or list of colors for plotting. If not set, clustering is done
        and colors are assigned automatically

    ax: matplotlib.Axes, optional
        Axes object to use.

    size: tuple, optional
        see waterfall

    x_label: str
        label for x-axis

    y_label: str
        label for y-axis

    legend_text: str
        Label for line plots

    Returns
    -------

    ax: matplotlib.Axes
        The plot axes.
    """

    # axes
    if ax is None:
        ax = plt.subplots()[1]
        fig = plt.gcf()
        fig.set_size_inches(*size)

    # parse input
    fvals = []
    if isinstance(vals, list):
        # convert entries to numpy arrays
        for val in vals:
            val = np.array(val)
            fvals.append(val[1, -1])
    else:
        # convert to a list of numpy arrays
        vals = np.array(vals)
        if vals.shape[0] != 2 or vals.ndim != 2:
            raise ValueError('If numpy array is passed directly to lowlevel '
                             'routine of optimizer_history, shape needs to '
                             'be 2 x n.')
        fvals = [vals[1, -1]]
        vals = [vals]
    n_fvals = len(fvals)

    # assign colors
    # note: this has to happen before sorting
    # to get the same colors in different plots
    colors = assign_colors(fvals, colors)

    # sort
    indices = sorted(range(n_fvals), key=lambda j: fvals[j])

    # plot
    ax.xaxis.set_major_locator(MaxNLocator(integer=True))
    for j, val in enumerate(vals):
        # collect and parse data
        j_fval = indices[j]
        color = colors[j_fval]
        if j == 0:
            tmp_legend = legend_text
        else:
            tmp_legend = None

        # line plots
        if scale_y == 'log10':
            ax.semilogy(val[0, :], val[1, :], color=color, label=tmp_legend)
        else:
            ax.plot(val[0, :], val[1, :], color=color, label=tmp_legend)

    # set labels
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)
    ax.set_title('Optimizer history')
    if legend_text is not None:
        ax.legend()

    return ax


def get_trace(result: Result,
              trace_x: Optional[str],
              trace_y: Optional[str]) -> Tuple[str, str, List[np.ndarray]]:
    """
    Get the values of the optimizer trace from the pypesto.Result object

    Parameters
    ----------

    result: pypesto.Result
        Optimization result obtained by 'optimize.py'.

    trace_x: str, optional
        What should be plotted on the x-axis?
        Possibilities: 'time', 'steps'
        Default: 'steps'

    trace_y: str, optional
        What should be plotted on the y-axis?
        Possibilities: 'fval'(later also: 'gradnorm', 'stepsize')
        Default: 'fval'

    Returns
    -------

    vals:
        list of

    x_label:
        label for x-axis to be plotted later

    y_label:
        label for y-axis to be plotted later
    """

    # get data frames
    histories: List[History] = result.optimize_result.get_for_key('history')

    vals = []
    x_label = ''
    y_label = ''

    for history in histories:
        options = history.options
        if trace_y == 'gradnorm':
            # retrieve gradient trace, if saved
            if not options.trace_record or not \
                    options.trace_record_grad:
                raise ValueError("No gradient trace has been recorded.")
            grads = history.get_grad_trace()
            indices = [i for i, val in enumerate(grads)
                       if val is not None and np.isfinite(val).all()]

            grads = np.array([grads[i] for i in indices])

            # Get gradient trace, compute norm
            y_vals = np.linalg.norm(grads, axis=1)
            y_label = 'gradient norm'

        else:  # trace_y == 'fval':
            if not options.trace_record:
                raise ValueError("No function value trace has been recorded.")
            fvals = history.get_fval_trace()
            indices = [i for i, val in enumerate(fvals)
                       if val is not None and np.isfinite(val)]

            y_vals = np.array([fvals[i] for i in indices])
            y_label = 'objective value'

        # retrieve values from dataframe
        if trace_x == 'time':
            times = np.array(history.get_time_trace())
            x_vals = times[indices]
            x_label = 'Computation time [s]'

        else:  # trace_x == 'steps':
            x_vals = np.array(list(range(len(indices))))
            x_label = 'Optimizer steps'

        # write down values
        vals.append(np.vstack([x_vals, y_vals]))

    return x_label, y_label, vals


def get_vals(
    vals: List[np.ndarray],
    scale_y: Optional[str],
    offset_y: float,
    y_label: str,
    start_indices: Iterable[int]
) -> Tuple[List[np.ndarray], float, str]:
    """
    Postprocesses the values of the optimization history, depending on the
    options set by the user (e.g. scale_y, offset_y, start_indices)

    Parameters
    ----------

    vals: list
        list of numpy arrays of dimension 2 x len(start_indices)

    scale_y: str, optional
        May be logarithmic or linear ('log10' or 'lin')

    offset_y: float
        offset for the y-axis, as this is supposed to be in log10-scale

    y_label: str
        Label for y axis

    start_indices:
        list of integers specifying the multi start indices to be plotted

    Returns
    -------

    vals: list
        list of numpy arrays

    offset_y:
        offset for the y-axis, if this is supposed to be in log10-scale

    y_label:
        Label for y axis
    """

    # get list of indices
    if start_indices is None:
        start_indices = np.array(range(len(vals)))
    else:
        # check whether list or maximum value
        start_indices = np.array(start_indices)

        # check, whether index set is not too big
        existing_indices = np.array(range(len(vals)))
        start_indices = np.intersect1d(start_indices, existing_indices)

    # reduce values to listed values
    vals = [val for i, val in enumerate(vals) if i in start_indices]

    # get the minimal value which should be plotted
    min_val = np.inf
    for val in vals:
        tmp_min = np.min(val[1, :])
        min_val = np.min([min_val, tmp_min])

    # check, whether offset can be used with this data
    if y_label == 'fval':
        offset_y = process_offset_y(offset_y, scale_y, min_val)
    else:
        offset_y = 0.

    if offset_y != 0:
        for val in vals:
            val[1, :] += offset_y * np.ones(val[1].shape)
            y_label = 'offsetted ' + y_label

    return vals, offset_y, y_label


def handle_options(ax: plt.Axes,
                   vals: List[np.ndarray],
                   ref: List[ReferencePoint],
                   y_limits: Union[float, np.ndarray],
                   offset_y: float):
    """
    Get the limits for the y-axis, plots the reference points, will do
    more at a later time point. This function is there to apply whatever
    kind of post-plotting transformations to the axis object.

    Parameters
    ----------

    ref:
        List of reference points for optimization results, containing et
        least a function value fval

    vals:
        list of numpy arrays of size 2 x number of values

    ax:
        Axes object to use.

    y_limits:
        maximum value to be plotted on the y-axis, or y-limits

    offset_y:
        offset for the y-axis, if this is supposed to be in log10-scale

    Returns
    -------

    ax: matplotlib.Axes
        The plot axes.
    """

    # handle y-limits
    ax = process_y_limits(ax, y_limits)

    # handle reference points
    if len(ref) > 0:
        # plot reference points
        # get length of longest trajectory
        max_len = 0
        for val in vals:
            max_len = np.max([max_len, val[0, -1]])

        for i_ref in ref:
            ax.plot([0, max_len],
                    [i_ref.fval + offset_y, i_ref.fval + offset_y],
                    '--', color=i_ref.color, label=i_ref.legend)

            # create legend for reference points
            if i_ref.legend is not None:
                ax.legend()

    return ax



import numpy as np

from ..objective import Objective
from ..problem import Problem

try:
    import theano.tensor as tt
    try:
        from theano.graph.null_type import NullType
    except ImportError:
        # for older versions of theano
        from theano.gof.null_type import NullType
except ImportError:
    tt = NullType = None


class TheanoLogProbability(tt.Op):
    """
    Theano wrapper around a (non-normalized) log-probability function.

    Parameters
    ----------
    problem:
        The `pypesto.Problem` to analyze.
    beta:
        Inverse temperature (e.g. in parallel tempering).
    """

    itypes = [tt.dvector]  # expects a vector of parameter values when called
    otypes = [tt.dscalar]  # outputs a single scalar value (the log prob)

    def __init__(self, problem: Problem, beta: float = 1.):
        self._objective: Objective = problem.objective

        # initialize the log probability Op
        self._log_prob = \
            lambda x: - beta * self._objective(x, sensi_orders=(0,))

        # initialize the sensitivity Op
        if problem.objective.has_grad:
            self._log_prob_grad = TheanoLogProbabilityGradient(problem, beta)
        else:
            self._log_prob_grad = None

    def perform(self, node, inputs, outputs, params=None):
        theta, = inputs
        log_prob = self._log_prob(theta)
        outputs[0][0] = np.array(log_prob)

    def grad(self, inputs, g):
        # the method that calculates the gradients - it actually returns the
        # vector-Jacobian product - g[0] is a vector of parameter values
        if self._log_prob_grad is None:
            # indicates gradient not available
            return [NullType]
        theta, = inputs
        log_prob_grad = self._log_prob_grad(theta)
        return [g[0] * log_prob_grad]


class TheanoLogProbabilityGradient(tt.Op):
    """
    Theano wrapper around a (non-normalized) log-probability gradient function.
    This Op will be called with a vector of values and also return a vector of
    values - the gradients in each dimension.

    Parameters
    ----------
    problem:
        The `pypesto.Problem` to analyze.
    beta:
        Inverse temperature (e.g. in parallel tempering).
    """

    itypes = [tt.dvector]  # expects a vector of parameter values when called
    otypes = [tt.dvector]  # outputs a vector (the log prob grad)

    def __init__(self, problem: Problem, beta: float = 1.):
        self._objective: Objective = problem.objective
        self._log_prob_grad = \
            lambda x: - beta * self._objective(x, sensi_orders=(1,))

    def perform(self, node, inputs, outputs, params=None):
        theta, = inputs
        # calculate gradients
        log_prob_grad = self._log_prob_grad(theta)
        outputs[0][0] = log_prob_grad



import numpy as np

from copy import deepcopy
from typing import Sequence, Tuple
from .base import ObjectiveBase, ResultDict

from .constants import RDATAS, FVAL, CHI2, SCHI2, RES, SRES, GRAD, HESS, HESSP


class AggregatedObjective(ObjectiveBase):
    """
    This class aggregates multiple objectives into one objective.
    """

    def __init__(
        self,
        objectives: Sequence[ObjectiveBase],
        x_names: Sequence[str] = None,
    ):
        """
        Constructor.


        Parameters
        ----------
        objectives:
            Sequence of pypesto.ObjectiveBase instances
        x_names:
            Sequence of names of the (optimized) parameters.
            (Details see documentation of x_names in
            :class:`pypesto.ObjectiveBase`)
        """

        # input typechecks
        if not isinstance(objectives, Sequence):
            raise TypeError(f'Objectives must be a Sequence, '
                            f'was {type(objectives)}.')

        if not all(
                isinstance(objective, ObjectiveBase)
                for objective in objectives
        ):
            raise TypeError('Objectives must only contain elements of type'
                            'pypesto.Objective')

        if not objectives:
            raise ValueError('Length of objectives must be at least one')

        self._objectives = objectives

        super().__init__(x_names=x_names)

    def __deepcopy__(self, memodict=None):
        other = AggregatedObjective(
            objectives=[deepcopy(objective) for objective in self._objectives],
            x_names=deepcopy(self.x_names),
        )
        for key in set(self.__dict__.keys()) - {'_objectives', 'x_names'}:
            other.__dict__[key] = deepcopy(self.__dict__[key])

        return other

    def check_mode(self, mode: str) -> bool:
        return all(
            objective.check_mode(mode)
            for objective in self._objectives
        )

    def check_sensi_orders(
        self,
        sensi_orders: Tuple[int, ...],
        mode: str,
    ) -> bool:
        return all(
            objective.check_sensi_orders(sensi_orders, mode)
            for objective in self._objectives
        )

    def call_unprocessed(
        self,
        x: np.ndarray,
        sensi_orders: Tuple[int, ...],
        mode: str,
        **kwargs,
    ) -> ResultDict:
        return aggregate_results([
            objective.call_unprocessed(x, sensi_orders, mode, **kwargs)
            for objective in self._objectives
        ])

    def initialize(self):
        for objective in self._objectives:
            objective.initialize()

    def get_config(self) -> dict:
        info = super().get_config()
        for n_obj, obj in enumerate(self._objectives):
            info[f'objective_{n_obj}'] = obj.get_config()
        return info


def aggregate_results(rvals: Sequence[ResultDict]) -> ResultDict:
    """
    Aggregrate the results from the provided sequence of ResultDicts into a
    single ResultDict.

    Parameters
    ----------
    rvals:
        results to aggregate
    """

    # rvals are guaranteed to be consistent as _check_sensi_orders checks
    # whether each objective can be called with the respective
    # sensi_orders/mode

    # sum over fval/grad/hess
    result = {
        key: sum(rval[key] for rval in rvals)
        for key in [FVAL, CHI2, SCHI2, GRAD, HESS, HESSP]
        if rvals[0].get(key, None) is not None
    }

    # extract rdatas and flatten
    result[RDATAS] = []
    for rval in rvals:
        if RDATAS in rval:
            result[RDATAS].extend(rval[RDATAS])

    # initialize res and sres
    if RES in rvals[0]:
        res = np.asarray(rvals[0][RES])
    else:
        res = None

    if SRES in rvals[0]:
        sres = np.asarray(rvals[0][SRES])
    else:
        sres = None

    # skip iobj=0 after initialization, stack matrices
    for rval in rvals[1:]:
        if res is not None:
            res = np.hstack([res, np.asarray(rval[RES])])
        if sres is not None:
            sres = np.vstack([sres, np.asarray(rval[SRES])])

    # fill res, sres into result
    if res is not None:
        result[RES] = res
    if sres is not None:
        result[SRES] = sres

    return result



"""
PEtab
=====

pyPESTO support for the PEtab data format.
"""
from .importer import PetabImporter
from .pysb_importer import PetabImporterPysb

# PEtab and amici are optional dependencies

import warnings

try:
    import petab
except ImportError:
    warnings.warn("PEtab import requires an installation of petab "
                  "(https://github.com/PEtab-dev/PEtab). "
                  "Install via `pip3 install petab`.")
try:
    import amici
except ImportError:
    warnings.warn("PEtab import requires an installation of amici "
                  "(https://github.com/AMICI-dev/AMICI). "
                  "Install via `pip3 install amici`.")



import numpy as np
from typing import Union, Tuple

from ..ensemble import Ensemble, EnsemblePrediction
from .utils import get_prediction_dataset


def get_covariance_matrix_parameters(ens: Ensemble) -> np.ndarray:
    """
    Compute the covariance of ensemble parameters.

    Parameters
    ==========
    ens:
        Ensemble object containing a set of parameter vectors

    Returns
    =======
    covariance_matrix:
        covariance matrix of ensemble parameters
    """

    # call lowlevel routine using the parameter ensemble
    return np.cov(ens.x_vectors.transpose())


def get_covariance_matrix_predictions(
        ens: Union[Ensemble, EnsemblePrediction],
        prediction_index: int = 0) -> np.ndarray:
    """
    Compute the covariance of ensemble predictions.

    Parameters
    ==========
    ens:
        Ensemble object containing a set of parameter vectors and a set of
        predictions or EnsemblePrediction object containing only predictions

    prediction_index:
        index telling which prediction from the list should be analyzed

    Returns
    =======
    covariance_matrix:
        covariance matrix of ensemble predictions
    """

    # extract the an array of predictions from either an Ensemble object or an
    # EnsemblePrediction object
    dataset = get_prediction_dataset(ens, prediction_index)

    # call lowlevel routine using the prediction ensemble
    return np.cov(dataset)


def get_spectral_decomposition_parameters(
        ens: Ensemble,
        normalize: bool = False,
        only_separable_directions: bool = False,
        cutoff_absolute_separable: float = 1e-16,
        cutoff_relative_separable: float = 1e-16,
        only_identifiable_directions: bool = False,
        cutoff_absolute_identifiable: float = 1e-16,
        cutoff_relative_identifiable: float = 1e-16
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Compute the spectral decomposition of ensemble parameters.

    Parameters
    ==========
    ens:
        Ensemble object containing a set of parameter vectors

    normalize:
        flag indicating whether the returned Eigenvalues should be normalized
        with respect to the largest Eigenvalue

    only_separable_directions:
        return only separable directions according to
        cutoff_[absolute/relative]_separable

    cutoff_absolute_separable:
        Consider only eigenvalues of the covariance matrix above this cutoff
        (only applied when only_separable_directions is True)

    cutoff_relative_separable:
        Consider only eigenvalues of the covariance matrix above this cutoff,
        when rescaled with the largest eigenvalue
        (only applied when only_separable_directions is True)

    only_identifiable_directions:
        return only identifiable directions according to
        cutoff_[absolute/relative]_identifiable

    cutoff_absolute_identifiable:
        Consider only low eigenvalues of the covariance matrix with inverses
        above of this cutoff
        (only applied when only_identifiable_directions is True)

    cutoff_relative_identifiable:
        Consider only low eigenvalues of the covariance matrix when rescaled
        with the largest eigenvalue with inverses above of this cutoff
        (only applied when only_identifiable_directions is True)

    Returns
    =======
    eigenvalues:
        Eigenvalues of the covariance matrix

    eigenvectors:
        Eigenvectors of the covariance matrix
    """
    # check inputs
    if sum([only_identifiable_directions, only_separable_directions]) >= 2:
        raise AssertionError(
            "Specify either only identifiable or only separable directions.")

    covariance = get_covariance_matrix_parameters(ens)
    return get_spectral_decomposition_lowlevel(
        matrix=covariance, normalize=normalize,
        only_separable_directions=only_separable_directions,
        cutoff_absolute_separable=cutoff_absolute_separable,
        cutoff_relative_separable=cutoff_relative_separable,
        only_identifiable_directions=only_identifiable_directions,
        cutoff_absolute_identifiable=cutoff_absolute_identifiable,
        cutoff_relative_identifiable=cutoff_relative_identifiable)


def get_spectral_decomposition_predictions(
        ens: Ensemble,
        normalize: bool = False,
        only_separable_directions: bool = False,
        cutoff_absolute_separable: float = 1e-16,
        cutoff_relative_separable: float = 1e-16,
        only_identifiable_directions: bool = False,
        cutoff_absolute_identifiable: float = 1e-16,
        cutoff_relative_identifiable: float = 1e-16
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Compute the spectral decomposition of ensemble predictions.

    Parameters
    ==========
    ens:
        Ensemble object containing a set of parameter vectors and a set of
        predictions or EnsemblePrediction object containing only predictions

    normalize:
        flag indicating whether the returned Eigenvalues should be normalized
        with respect to the largest Eigenvalue

    only_separable_directions:
        return only separable directions according to
        cutoff_[absolute/relative]_separable

    cutoff_absolute_separable:
        Consider only eigenvalues of the covariance matrix above this cutoff
        (only applied when only_separable_directions is True)

    cutoff_relative_separable:
        Consider only eigenvalues of the covariance matrix above this cutoff,
        when rescaled with the largest eigenvalue
        (only applied when only_separable_directions is True)

    only_identifiable_directions:
        return only identifiable directions according to
        cutoff_[absolute/relative]_identifiable

    cutoff_absolute_identifiable:
        Consider only low eigenvalues of the covariance matrix with inverses
        above of this cutoff
        (only applied when only_identifiable_directions is True)

    cutoff_relative_identifiable:
        Consider only low eigenvalues of the covariance matrix when rescaled
        with the largest eigenvalue with inverses above of this cutoff
        (only applied when only_identifiable_directions is True)

    Returns
    =======
    eigenvalues:
        Eigenvalues of the covariance matrix

    eigenvectors:
        Eigenvectors of the covariance matrix
    """
    covariance = get_covariance_matrix_predictions(ens)
    return get_spectral_decomposition_lowlevel(
        matrix=covariance, normalize=normalize,
        only_separable_directions=only_separable_directions,
        cutoff_absolute_separable=cutoff_absolute_separable,
        cutoff_relative_separable=cutoff_relative_separable,
        only_identifiable_directions=only_identifiable_directions,
        cutoff_absolute_identifiable=cutoff_absolute_identifiable,
        cutoff_relative_identifiable=cutoff_relative_identifiable)


def get_spectral_decomposition_lowlevel(
        matrix: np.ndarray,
        normalize: bool = False,
        only_separable_directions: bool = False,
        cutoff_absolute_separable: float = 1e-16,
        cutoff_relative_separable: float = 1e-16,
        only_identifiable_directions: bool = False,
        cutoff_absolute_identifiable: float = 1e-16,
        cutoff_relative_identifiable: float = 1e-16
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Compute the spectral decomposition of ensemble parameters or predictions.

    Parameters
    ==========
    matrix:
        symmetric matrix (typically a covariance matrix) of parameters or
        predictions

    normalize:
        flag indicating whether the returned Eigenvalues should be normalized
        with respect to the largest Eigenvalue

    only_separable_directions:
        return only separable directions according to
        cutoff_[absolute/relative]_separable

    cutoff_absolute_separable:
        Consider only eigenvalues of the covariance matrix above this cutoff
        (only applied when only_separable_directions is True)

    cutoff_relative_separable:
        Consider only eigenvalues of the covariance matrix above this cutoff,
        when rescaled with the largest eigenvalue
        (only applied when only_separable_directions is True)

    only_identifiable_directions:
        return only identifiable directions according to
        cutoff_[absolute/relative]_identifiable

    cutoff_absolute_identifiable:
        Consider only low eigenvalues of the covariance matrix with inverses
        above of this cutoff
        (only applied when only_identifiable_directions is True)

    cutoff_relative_identifiable:
        Consider only low eigenvalues of the covariance matrix when rescaled
        with the largest eigenvalue with inverses above of this cutoff
        (only applied when only_identifiable_directions is True)

    Returns
    =======
    eigenvalues:
        Eigenvalues of the covariance matrix

    eigenvectors:
        Eigenvectors of the covariance matrix
    """

    # get the eigenvalue decomposition
    eigenvalues, eigenvectors = np.linalg.eigh(matrix)

    # get a normalized version
    rel_eigenvalues = eigenvalues / np.max(eigenvalues)

    # If no filtering is wanted, we can return
    if not only_identifiable_directions and not only_separable_directions:
        # apply normalization
        if normalize:
            eigenvalues = rel_eigenvalues
        return eigenvalues, eigenvectors

    # Separable directions are wanted: an upper pass filtering is needed
    if only_separable_directions:
        if cutoff_absolute_separable is not None and \
                cutoff_relative_separable is not None:
            above_cutoff = np.array([
                i_eig_abs > cutoff_absolute_separable and
                i_eig_rel > cutoff_relative_separable
                for i_eig_abs, i_eig_rel in zip(eigenvalues, rel_eigenvalues)
            ])
        elif cutoff_absolute_separable is not None:
            above_cutoff = eigenvalues > cutoff_absolute_separable
        elif cutoff_relative_separable is not None:
            above_cutoff = rel_eigenvalues > cutoff_relative_separable
        else:
            raise Exception('Need a lower cutoff (absolute or relative, '
                            'e.g., 1e-16, to compute separable directions.')

        # restrict to those above cutoff
        eigenvalues = eigenvalues[above_cutoff]
        eigenvectors = eigenvectors[:, above_cutoff]
        # apply normlization
        if normalize:
            eigenvalues = rel_eigenvalues[above_cutoff]
        return eigenvalues, eigenvectors

    # Identifiable directions are wanted: an filtering of the inverse
    # eigenvalues is needed (upper pass of inverse = lower pass of original)
    if cutoff_absolute_identifiable is not None and \
            cutoff_relative_identifiable is not None:
        below_cutoff = np.array([
            1 / i_eig_abs > cutoff_absolute_identifiable and
            1 / i_eig_rel > cutoff_relative_identifiable
            for i_eig_abs, i_eig_rel in zip(eigenvalues, rel_eigenvalues)
        ])
    elif cutoff_absolute_identifiable is not None:
        below_cutoff = 1 / eigenvalues > cutoff_absolute_identifiable
    elif cutoff_relative_identifiable is not None:
        below_cutoff = 1 / rel_eigenvalues > cutoff_relative_identifiable
    else:
        raise Exception('Need an inverse upper cutoff (absolute or relative, '
                        'e.g., 1e-16, to compute identifiable directions.')

    # restrict to those below cutoff
    eigenvalues = eigenvalues[below_cutoff]
    eigenvectors = eigenvectors[:, below_cutoff]
    # apply normlization
    if normalize:
        eigenvalues = rel_eigenvalues[below_cutoff]
    return eigenvalues, eigenvectors



"""
Startpoint
==========

Methods for selecting points that can be used as startpoints
for multi-start optimization.
Startpoint methods can be implemented by deriving from
:class:`pypesto.startpoint.StartpointMethod`.
Handling in pypesto is then wrapped in
:func:`pypesto.startpoint.assign_startpoints`, handling e.g.
non-requirement of startpoints e.g. for global methods, and re-sampling of
non-finite points.
"""

from .base import StartpointMethod, FunctionStartpoints, to_startpoint_method
from .uniform import UniformStartpoints, uniform
from .latin_hypercube import LatinHypercubeStartpoints, latin_hypercube
from .assign import assign_startpoints



from typing import List, Union
import numpy as np

from ..problem import Problem
from .sampler import Sampler
from .result import McmcPtResult
from ..startpoint import assign_startpoints, uniform

try:
    import emcee
except ImportError:
    emcee = None


class EmceeSampler(Sampler):
    """Use emcee for sampling.

    Wrapper around https://emcee.readthedocs.io/en/stable/, see there for
    details.
    """

    def __init__(
        self,
        nwalkers: int = 1,
        sampler_args: dict = None,
        run_args: dict = None,
    ):
        """
        Parameters
        ----------
        nwalkers: The number of walkers in the ensemble.
        sampler_args:
            Further keyword arguments that are passed on to
            ``emcee.EnsembleSampler.__init__``.
        run_args:
            Further keyword arguments that are passed on to
            ``emcee.EnsembleSampler.run_mcmc``.
        """
        # check dependencies
        if emcee is None:
            raise ImportError(
                "This sampler requires an installation of emcee. Install e.g. "
                "via ``pip install pypesto[emcee]``.")

        super().__init__()
        self.nwalkers: int = nwalkers

        if sampler_args is None:
            sampler_args = {}
        self.sampler_args: dict = sampler_args

        if run_args is None:
            run_args = {}
        self.run_args: dict = run_args

        # set in initialize
        self.problem: Union[Problem, None] = None
        self.sampler: Union[emcee.EnsembleSampler, None] = None
        self.state: Union[emcee.State, None] = None

    def initialize(
        self,
        problem: Problem,
        x0: Union[np.ndarray, List[np.ndarray]],
    ) -> None:
        self.problem = problem

        # extract for pickling efficiency
        objective = self.problem.objective
        lb = self.problem.lb
        ub = self.problem.ub

        # parameter dimenstion
        ndim = len(self.problem.x_free_indices)

        def log_prob(x):
            """Log-probability density function."""
            # check if parameter lies within bounds
            if any(x < lb) or any(x > ub):
                return - np.inf
            # invert sign
            return - 1. * objective(x)

        # initialize sampler
        self.sampler = emcee.EnsembleSampler(
            nwalkers=self.nwalkers,
            ndim=ndim,
            log_prob_fn=log_prob,
            **self.sampler_args,
        )

        # assign startpoints
        if self.state is None:
            #  extract x0
            x0 = np.asarray(x0)
            if x0.ndim == 1:
                x0 = [x0]
            x0 = np.array([problem.get_full_vector(x) for x in x0])
            #  add x0 to guesses
            problem.x_guesses_full = np.row_stack((x0, problem.x_guesses_full))

            #  sample start points
            self.state = assign_startpoints(
                n_starts=self.nwalkers, startpoint_method=uniform,
                problem=problem, startpoint_resample=True)

            #  restore original guesses
            problem.x_guesses_full = problem.x_guesses_full[x0.shape[0]:]

    def sample(self, n_samples: int, beta: float = 1.) -> None:
        # the method returns the most recent sample state
        self.state = self.sampler.run_mcmc(
            self.state, n_samples, **self.run_args)

    def get_samples(self) -> McmcPtResult:
        # all walkers are concatenated, yielding a flat array
        trace_x = np.array([self.sampler.get_chain(flat=True)])
        trace_neglogpost = np.array([- self.sampler.get_log_prob(flat=True)])
        # the sampler does not know priors
        trace_neglogprior = np.full(trace_neglogpost.shape, np.nan)
        # the walkers all run on temperature 1
        betas = np.array([1.])

        result = McmcPtResult(
            trace_x=trace_x,
            trace_neglogpost=trace_neglogpost,
            trace_neglogprior=trace_neglogprior,
            betas=betas,
        )

        return result



import numpy as np
from typing import Union
import logging

from ..objective import History
from ..problem import Problem
from .sampler import Sampler
from .result import McmcPtResult

logger = logging.getLogger(__name__)

try:
    import pymc3 as pm
    import arviz as az
    import theano.tensor as tt
except ImportError:
    pm = az = tt = None

try:
    from .theano import TheanoLogProbability
except (AttributeError, ImportError):
    TheanoLogProbability = None


class Pymc3Sampler(Sampler):
    """Wrapper around Pymc3 samplers.

    Parameters
    ----------
    step_function:
        A pymc3 step function, e.g. NUTS, Slice. If not specified, pymc3
        determines one automatically (preferable).
    **kwargs:
        Options are directly passed on to `pymc3.sample`.
    """

    def __init__(self, step_function=None, **kwargs):
        super().__init__(kwargs)
        self.step_function = step_function
        self.problem: Union[Problem, None] = None
        self.x0: Union[np.ndarray, None] = None
        self.trace: Union[pm.backends.Text, None] = None
        self.data: Union[az.InferenceData, None] = None

    @classmethod
    def translate_options(cls, options):
        if not options:
            options = {'chains': 1}
        return options

    def initialize(self, problem: Problem, x0: np.ndarray):
        self.problem = problem
        if x0 is not None:
            if len(x0) != problem.dim:
                x0 = problem.get_reduced_vector(x0)
        self.x0 = x0
        self.trace = None
        self.data = None

        self.problem.objective.history = History()

    def sample(
            self, n_samples: int, beta: float = 1.):
        problem = self.problem
        log_post_fun = TheanoLogProbability(problem, beta)
        trace = self.trace

        x0 = None
        if self.x0 is not None:
            x0 = {x_name: val
                  for x_name, val in zip(self.problem.x_names, self.x0)}

        # create model context
        with pm.Model() as model:
            # uniform bounds
            k = [pm.Uniform(x_name, lower=lb, upper=ub)
                 for x_name, lb, ub in
                 zip(problem.get_reduced_vector(problem.x_names),
                     problem.lb, problem.ub)]

            # convert to tensor vector
            theta = tt.as_tensor_variable(k)

            # use a DensityDist for the log-posterior
            pm.DensityDist('log_post', logp=lambda v: log_post_fun(v),
                           observed={'v': theta})

            # step, by default automatically determined by pymc3
            step = None
            if self.step_function:
                step = self.step_function()

            # perform the actual sampling
            trace = pm.sample(
                draws=int(n_samples), trace=trace, start=x0, step=step,
                **self.options)

            # convert trace to inference data object
            data = az.from_pymc3(trace=trace, model=model)

        self.trace = trace
        self.data = data

    def get_samples(self) -> McmcPtResult:
        # parameter values
        trace_x = np.asarray(
            self.data.posterior.to_array()).transpose((1, 2, 0))

        # TODO this is only the negative objective values
        trace_neglogpost = np.asarray(self.data.log_likelihood.to_array())
        # remove trailing dimensions
        trace_neglogpost = np.reshape(trace_neglogpost,
                                      trace_neglogpost.shape[1:-1])
        # flip sign
        trace_neglogpost = - trace_neglogpost

        if trace_x.shape[0] != trace_neglogpost.shape[0] \
                or trace_x.shape[1] != trace_neglogpost.shape[1] \
                or trace_x.shape[2] != self.problem.dim:
            raise ValueError("Trace dimensions are inconsistent")

        return McmcPtResult(
            trace_x=np.array(trace_x),
            trace_neglogpost=np.array(trace_neglogpost),
            trace_neglogprior=np.full(trace_neglogpost.shape, np.nan),
            betas=np.array([1.] * trace_x.shape[0]),
        )



from .importer import PetabImporter
import os
import shutil

try:
    import amici
    import amici.petab_import_pysb
except ImportError:
    pass


class PetabImporterPysb(PetabImporter):
    """Import for experimental PySB-based PEtab problems"""

    def __init__(self,
                 petab_problem: 'amici.petab_import_pysb.PysbPetabProblem',
                 output_folder: str = None):
        """
        petab_problem:
            Managing access to the model and data.
        output_folder:
            Folder to contain the amici model.
        """

        super().__init__(petab_problem,
                         model_name=petab_problem.pysb_model.name,
                         output_folder=output_folder,
                         validate_petab=False)

    def compile_model(self, **kwargs):
        """
        Compile the model. If the output folder exists already, it is first
        deleted.

        Parameters
        ----------
        kwargs: Extra arguments passed to `amici.SbmlImporter.sbml2amici`.

        """

        # delete output directory
        if os.path.exists(self.output_folder):
            shutil.rmtree(self.output_folder)

        amici.petab_import_pysb.import_model_pysb(
            petab_problem=self.petab_problem,
            model_output_dir=self.output_folder,
            **kwargs)



import h5py
from ..result import Result
from ..optimize.result import OptimizerResult
from ..optimize.optimizer import fill_result_from_objective_history
from ..profile.result import ProfilerResult
from ..sample.result import McmcPtResult
from ..problem import Problem
from ..objective import Objective, ObjectiveBase, Hdf5History, OptimizerHistory
import numpy as np
import ast
import logging


logger = logging.getLogger(__name__)


def read_hdf5_profile(f: h5py.File,
                      profile_id: str,
                      parameter_id: str) -> 'ProfilerResult':
    """
    Read HDF5 results per start.

    Parameters
    -------------
    f:
        The HDF5 result file
    profile_id:
        specifies the profile start that is read
        from the HDF5 file
    parameter_id:
        specifies the profile index that is read
        from the HDF5 file
    """

    result = ProfilerResult(np.array([]), np.array([]), np.array([]))

    for profile_key in result.keys():
        if profile_key in f[f'/profiling/{profile_id}/{parameter_id}']:
            result[profile_key] = \
                f[f'/profiling/{profile_id}/{parameter_id}/{profile_key}'][:]
        elif profile_key in \
                f[f'/profiling/{profile_id}/{parameter_id}'].attrs:
            result[profile_key] = \
                f[f'/profiling/{profile_id}/{parameter_id}'].attrs[profile_key]
    return result


def read_hdf5_optimization(f: h5py.File,
                           file_name: str,
                           opt_id: str) -> 'OptimizerResult':
    """
    Read HDF5 results per start.

    Parameters
    -------------
    f:
        The HDF5 result file
    file_name:
        The name of the HDF5 file, needed to create HDF5History
    opt_id:
        Specifies the start that is read from the HDF5 file
    """

    result = OptimizerResult()

    for optimization_key in result.keys():
        if optimization_key == 'history':
            if optimization_key in f:
                result['history'] = Hdf5History(id=opt_id,
                                                file=file_name)
                result['history']._recover_options(file_name)
                continue
        if optimization_key in f[f'/optimization/results/{opt_id}']:
            result[optimization_key] = \
                f[f'/optimization/results/{opt_id}/{optimization_key}'][:]
        elif optimization_key in \
                f[f'/optimization/results/{opt_id}'].attrs:
            result[optimization_key] = \
                f[f'/optimization/results/{opt_id}'].attrs[optimization_key]
    return result


class ProblemHDF5Reader:
    """
    Reader of the HDF5 problem files written
    by class ProblemHDF5Writer.

    Attributes
    -------------
    storage_filename:
        HDF5 problem file name
    """
    def __init__(self, storage_filename: str):
        """
        Parameters
        ----------

        storage_filename: str
            HDF5 problem file name
        """
        self.storage_filename = storage_filename

    def read(self, objective: ObjectiveBase = None) -> Problem:
        """
        Read HDF5 problem file and return pyPESTO problem object.

        Parameters
        ----------
        objective:
            Objective function which is currently not saved to storage.
        Returns
        -------
        problem:
            A problem instance with all attributes read in.
        """

        # create empty problem
        if objective is None:
            objective = Objective()
            # raise warning that objective is not loaded.
            logger.info('WARNING: You are loading a problem.\nThis problem'
                        ' is not to be used without a separately created'
                        ' objective.')
        problem = Problem(objective, [], [])

        with h5py.File(self.storage_filename, 'r') as f:
            for problem_key in f['/problem']:
                if problem_key == 'config':
                    continue
                setattr(problem, problem_key,
                        f[f'/problem/{problem_key}'][:])
            for problem_attr in f['/problem'].attrs:
                setattr(problem, problem_attr,
                        f['/problem'].attrs[problem_attr])

        # h5 uses numpy for everything; convert to lists where necessary
        problem.x_fixed_vals = [float(val) for val in problem.x_fixed_vals]
        problem.x_fixed_indices = [int(ix) for ix in problem.x_fixed_indices]
        problem.x_names = [name.decode() for name in problem.x_names]

        return problem


class OptimizationResultHDF5Reader:
    """
    Reader of the HDF5 result files written
    by class OptimizationResultHDF5Writer.

    Attributes
    -------------
    storage_filename:
        HDF5 result file name
    """
    def __init__(self, storage_filename: str):
        """
        Parameters
        ----------

        storage_filename: str
            HDF5 result file name
        """
        self.storage_filename = storage_filename
        self.results = Result()

    def read(self) -> Result:
        """
        Read HDF5 result file and return pyPESTO result object.
        """
        with h5py.File(self.storage_filename, "r") as f:
            if '/problem' in f['/']:
                problem_reader = ProblemHDF5Reader(self.storage_filename)
                self.results.problem = problem_reader.read()

            for opt_id in f['/optimization/results']:
                result = read_hdf5_optimization(f,
                                                self.storage_filename,
                                                opt_id)
                self.results.optimize_result.append(result)
            self.results.optimize_result.sort()
        return self.results


class SamplingResultHDF5Reader:
    """
    Reader of the HDF5 result files written
    by class SamplingResultHDF5Writer.

    Attributes
    -------------
    storage_filename:
        HDF5 result file name
    """
    def __init__(self, storage_filename: str):
        """
        Parameters
        ----------

        storage_filename: str
            HDF5 result file name
        """
        self.storage_filename = storage_filename
        self.results = Result()

    def read(self) -> Result:
        """
        Read HDF5 result file and return pyPESTO result object.
        """
        sample_result = {}
        with h5py.File(self.storage_filename, "r") as f:
            if '/problem' in f['/']:
                problem_reader = ProblemHDF5Reader(self.storage_filename)
                self.results.problem = problem_reader.read()
            for key in f['/sampling/results']:
                sample_result[key] = \
                    f[f'/sampling/results/{key}'][:]
            for key in f['/sampling/results'].attrs:
                sample_result[key] = \
                    f['/sampling/results'].attrs[key]
        try:
            self.results.sample_result = McmcPtResult(**sample_result)
        except TypeError:
            logger.warning("Warning: You tried loading a non-existent "
                           "sampling result.")

        return self.results


class ProfileResultHDF5Reader:
    """
    Reader of the HDF5 result files written
    by class OptimizationResultHDF5Writer.

    Attributes
    -------------
    storage_filename:
        HDF5 result file name
    """
    def __init__(self, storage_filename: str):
        """
        Parameters
        ----------

        storage_filename: str
            HDF5 result file name
        """
        self.storage_filename = storage_filename
        self.results = Result()

    def read(self) -> Result:
        """
        Read HDF5 result file and return pyPESTO result object.
        """
        profiling_list = []
        with h5py.File(self.storage_filename, "r") as f:
            if '/problem' in f['/']:
                problem_reader = ProblemHDF5Reader(self.storage_filename)
                self.results.problem = problem_reader.read()
            for profile_id in f['/profiling']:
                profiling_list.append([
                    None for _ in f[f'/profiling/{profile_id}']
                ])
                for parameter_id in f[f'/profiling/{profile_id}']:
                    if f[f'/profiling/{profile_id}/'
                         f'{parameter_id}'].attrs['IsNone']:
                        continue
                    profiling_list[int(profile_id)][int(parameter_id)] = \
                        read_hdf5_profile(f,
                                          profile_id=profile_id,
                                          parameter_id=parameter_id)
            self.results.profile_result.list = profiling_list
        return self.results


def read_result(filename: str,
                problem: bool = True,
                optimize: bool = False,
                profile: bool = False,
                sample: bool = False,
                ) -> Result:
    """
    This is a function that saves the whole pypesto.Result object in an
    HDF5 file. With booleans one can choose more detailed what to save.

    Parameters
    ----------
    filename:
        The HDF5 filename.
    problem:
        Read the problem.
    optimize:
        Read the optimize result.
    profile:
        Read the profile result.
    sample:
        Read the sample result.

    Returns
    -------
    result:
        Result object containing the results stored in HDF5 file.
    """
    if not any([optimize, profile, sample]):
        optimize = True
        profile = True
        sample = True
    result = Result()

    if problem:
        pypesto_problem_reader = ProblemHDF5Reader(filename)
        result.problem = pypesto_problem_reader.read()

    if optimize:
        pypesto_opt_reader = OptimizationResultHDF5Reader(filename)
        try:
            temp_result = pypesto_opt_reader.read()
            result.optimize_result = temp_result.optimize_result
        except KeyError:
            logger.warning('Loading the optimization result failed. It is '
                           'highly likely that no optimization result exists '
                           f'within {filename}.')

    if profile:
        pypesto_profile_reader = ProfileResultHDF5Reader(filename)
        try:
            temp_result = pypesto_profile_reader.read()
            result.profile_result = temp_result.profile_result
        except KeyError:
            logger.warning('Loading the profiling result failed. It is '
                           'highly likely that no profiling result exists '
                           f'within {filename}.')

    if sample:
        pypesto_sample_reader = SamplingResultHDF5Reader(filename)
        try:
            temp_result = pypesto_sample_reader.read()
            result.sample_result = temp_result.sample_result
        except KeyError:
            logger.warning('Loading the sampling result failed. It is '
                           'highly likely that no sampling result exists '
                           f'within {filename}.')

    return result


def load_objective_config(filename: str):
    """
    Load the objective information stored in f

    Parameters
    ----------
    filename:
        The name of the file in which the information are stored.

    Returns:
        A dictionary of the information, stored instead of the
        actual objective in problem.objective.
    """

    with h5py.File(filename, 'r') as f:
        info_str = f['problem/config'][()].decode()
        info = ast.literal_eval(info_str)
        return info


def optimization_result_from_history(filename: str):
    """
    Converts a saved hdf5 History to an optimization result.
    Used for interrupted optimization runs.

    Parameters
    ----------
    filename:
        The name of the file in which the information are stored.

    Returns:
        A result object in which the optimization result is constructed from
        history. But missing "Time, Message and Exitflag" keys.
    """

    result = Result()
    with h5py.File(filename, 'r') as f:
        for id_name in f['history'].keys():
            history = Hdf5History(id=id_name, file=filename)
            history._recover_options(filename)
            optimizer_history = OptimizerHistory(
                history,
                x0=f[f'history/{id_name}/trace/0/x'][()],
                generate_from_history=True)
            optimizer_result = OptimizerResult(id=id_name)
            fill_result_from_objective_history(optimizer_result,
                                               optimizer_history)
            result.optimize_result.append(optimizer_result)
    return result



"""Startpoint base classes."""

import numpy as np
from abc import abstractmethod
from typing import Callable, Union

from ..objective import ObjectiveBase


class StartpointMethod:
    """Startpoint generation, in particular for multi-start optimization.

    Abstract base class, specific sampling method needs to be defined in
    sub-classes.
    """

    @abstractmethod
    def __call__(
        self,
        n_starts: int,
        lb: np.ndarray,
        ub: np.ndarray,
        objective: ObjectiveBase,
    ) -> np.ndarray:
        """Generate startpoints.

        Parameters
        ----------
        n_starts:
            Number of starts.
        lb:
            Lower parameter bound.
        ub:
            Upper parameter bound.
        objective:
            Objective, maybe required for evaluation.

        Returns
        -------
        startpoints:
            Startpoints, shape (n_starts, n_par).
        """


class FunctionStartpoints(StartpointMethod):
    """Define startpoints via callable.

    The callable should take the same arguments as the `__call__` method.
    """

    def __init__(
        self,
        function: Callable,
    ):
        """
        Parameters
        ----------
        function: The callable sampling startpoints.
        """
        self.function = function

    def __call__(
            self,
            n_starts: int,
            lb: np.ndarray,
            ub: np.ndarray,
            objective: ObjectiveBase,
    ) -> np.ndarray:
        return self.function(
            n_starts=n_starts, lb=lb, ub=ub, objective=objective,
        )


def to_startpoint_method(
    maybe_startpoint_method: Union[StartpointMethod, Callable],
) -> StartpointMethod:
    """Create StartpointMethod instance if possible, otherwise raise.

    Parameters
    ----------
    maybe_startpoint_method:
        A StartpointMethod instance, or a Callable as expected by
        FunctionStartpoints.

    Returns
    -------
    startpoint_method:
        A StartpointMethod instance.

    Raises
    ------
    TypeError if arguments cannot be converted to a StartpointMethod.
    """
    if isinstance(maybe_startpoint_method, StartpointMethod):
        return maybe_startpoint_method
    if isinstance(maybe_startpoint_method, Callable):
        return FunctionStartpoints(maybe_startpoint_method)
    raise TypeError(
        "Could not parse startpoint method of type "
        f"{type(maybe_startpoint_method)} to a StartpointMethod.",
    )



import matplotlib.pyplot as plt
import matplotlib.axes
from matplotlib.ticker import MaxNLocator
import numpy as np

from typing import Iterable, List, Optional, Sequence, Tuple, Union

from ..result import Result
from .reference_points import create_references, ReferencePoint
from .clust_color import assign_colors
from .clust_color import delete_nan_inf
from .misc import process_result_list, process_start_indices


def parameters(
        results: Union[Result, Sequence[Result]],
        ax: Optional[matplotlib.axes.Axes] = None,
        parameter_indices: Union[str, Sequence[int]] = 'free_only',
        lb: Optional[Union[np.ndarray, List[float]]] = None,
        ub: Optional[Union[np.ndarray, List[float]]] = None,
        size: Optional[Tuple[float, float]] = None,
        reference: Optional[List[ReferencePoint]] = None,
        colors: Optional[Union[List[float], List[List[float]]]] = None,
        legends: Optional[Union[str, List[str]]] = None,
        balance_alpha: bool = True,
        start_indices: Optional[Union[int, Iterable[int]]] = None,
        scale_to_interval: Optional[Tuple[float, float]] = None,
) -> matplotlib.axes.Axes:
    """
    Plot parameter values.

    Parameters
    ----------
    results:
        Optimization result obtained by 'optimize.py' or list of those
    ax:
        Axes object to use.
    parameter_indices:
        Specifies which parameters should be plotted. Allowed string values are
        'all' (both fixed and free parameters will be plotted)  and
        'free_only' (only free parameters will be plotted)
    lb, ub:
        If not None, override result.problem.lb, problem.problem.ub.
        Dimension either result.problem.dim or result.problem.dim_full.
    size:
        Figure size (width, height) in inches. Is only applied when no ax
        object is specified
    reference:
        List of reference points for optimization results, containing at
        least a function value fval
    colors:
        list of RGBA colors, or single RGBA color
        If not set, clustering is done and colors are assigned automatically
    legends:
        Labels for line plots, one label per result object
    balance_alpha:
        Flag indicating whether alpha for large clusters should be reduced to
        avoid overplotting (default: True)
    start_indices:
        list of integers specifying the multistarts to be plotted or
        int specifying up to which start index should be plotted
    scale_to_interval:
        Tuple of bounds to which to scale all parameter values and bounds, or
        ``None`` to use bounds as determined by ``lb, ub``.

    Returns
    -------
    ax:
        The plot axes.
    """

    # parse input
    (results, colors, legends) = process_result_list(results, colors, legends)

    if isinstance(parameter_indices, str):
        if parameter_indices == 'all':
            parameter_indices = range(0, results[0].problem.dim_full)
        elif parameter_indices == 'free_only':
            parameter_indices = results[0].problem.x_free_indices
        else:
            raise ValueError("Permissible values for parameter_indices are "
                             "'all', 'free_only' or a list of indices")

    def scale_parameters(x):
        """Scale ``x`` from [lb, ub] to interval given by ``scale_to_interval``
        """
        if scale_to_interval is None or scale_to_interval is False:
            return x

        return (scale_to_interval[0] + (x - lb) / (ub - lb)
                * (scale_to_interval[1] - scale_to_interval[0]))

    for j, result in enumerate(results):
        # handle results and bounds
        (lb, ub, x_labels, fvals, xs) = \
            handle_inputs(result=result, lb=lb, ub=ub,
                          parameter_indices=parameter_indices,
                          start_indices=start_indices)
        lb, ub, xs = map(scale_parameters, (lb, ub, xs))

        # call lowlevel routine
        ax = parameters_lowlevel(xs=xs, fvals=fvals, lb=lb, ub=ub,
                                 x_labels=x_labels, ax=ax, size=size,
                                 colors=colors[j], legend_text=legends[j],
                                 balance_alpha=balance_alpha)

    # parse and apply plotting options
    ref = create_references(references=reference)

    # plot reference points
    for i_ref in ref:
        # reduce parameter vector in reference point, if necessary
        if len(parameter_indices) < results[0].problem.dim_full:
            x_ref = np.array(results[0].problem.get_reduced_vector(
                i_ref['x'],
                parameter_indices))
        else:
            x_ref = np.array(i_ref['x'])
        x_ref = np.reshape(x_ref, (1, x_ref.size))
        x_ref = scale_parameters(x_ref)

        # plot reference parameters using lowlevel routine
        ax = parameters_lowlevel(x_ref, [i_ref['fval']], ax=ax,
                                 colors=i_ref['color'],
                                 linestyle='--',
                                 legend_text=i_ref.legend,
                                 balance_alpha=balance_alpha)

    return ax


def parameter_hist(
        result: Result,
        parameter_name: str,
        bins: Union[int, str] = 'auto',
        ax: Optional['matplotlib.Axes'] = None,
        size: Optional[Tuple[float]] = (18.5, 10.5),
        color: Optional[List[float]] = None,
        start_indices: Optional[Union[int, List[int]]] = None):
    """
    Plot parameter values as a histogram.

    Parameters
    ----------
    result:
        Optimization result obtained by 'optimize.py'
    parameter_name:
        The name of the parameter that should be plotted
    bins:
        Specifies bins of the histogram
    ax:
        Axes object to use
    size:
        Figure size (width, height) in inches. Is only applied when no ax
        object is specified
    color:
        RGBA color.
    start_indices:
        List of integers specifying the multistarts to be plotted or
        int specifying up to which start index should be plotted


    Returns
    -------
    ax:
    The plot axes.

    """

    if ax is None:
        ax = plt.subplots()[1]
        fig = plt.gcf()
        fig.set_size_inches(*size)

    xs = result.optimize_result.get_for_key('x')

    # reduce number of displayed results
    if isinstance(start_indices, int):
        xs = xs[:start_indices]
    elif start_indices is not None:
        xs = [xs[ind] for ind in start_indices]

    parameter_index = result.problem.x_names.index(parameter_name)
    parameter_values = [x[parameter_index] for x in xs]

    ax.hist(parameter_values, color=color, bins=bins)
    ax.set_xlabel(parameter_name)
    ax.set_ylabel("counts")
    ax.set_title(f"Parameter {parameter_name}")

    return ax


def parameters_lowlevel(
        xs: Sequence[Union[np.ndarray, List[float]]],
        fvals: Union[np.ndarray, List[float]],
        lb: Optional[Union[np.ndarray, List[float]]] = None,
        ub: Optional[Union[np.ndarray, List[float]]] = None,
        x_labels: Optional[Iterable[str]] = None,
        ax: Optional[matplotlib.axes.Axes] = None,
        size: Optional[Tuple[float, float]] = None,
        colors: Optional[Sequence[Union[np.ndarray, List[float]]]] = None,
        linestyle: str = '-',
        legend_text: Optional[str] = None,
        balance_alpha: bool = True
) -> matplotlib.axes.Axes:

    """
    Plot parameters plot using list of parameters.

    Parameters
    ----------

    xs:
        Including optimized parameters for each startpoint.
        Shape: (n_starts, dim).

    fvals:
        Function values. Needed to assign cluster colors.

    lb, ub:
        The lower and upper bounds.

    x_labels:
        Labels to be used for the parameters.

    ax:
        Axes object to use.

    size:
        see parameters

    colors:
        One for each element in 'fvals'.

    linestyle:
        linestyle argument for parameter plot

    legend_text:
        Label for line plots

    balance_alpha:
        Flag indicating whether alpha for large clusters should be reduced to
        avoid overplotting (default: True)

    Returns
    -------

    ax:
        The plot axes.
    """

    # parse input
    xs = np.array(xs)
    fvals = np.array(fvals)
    # remove nan or inf values in fvals and xs
    xs, fvals = delete_nan_inf(fvals, xs, len(ub) if ub is not None else 1)

    if size is None:
        # 0.5 inch height per parameter
        size = (18.5, max(xs.shape[1], 1) / 2)

    if ax is None:
        ax = plt.subplots()[1]
        fig = plt.gcf()
        fig.set_size_inches(*size)

    # assign colors
    colors = assign_colors(vals=fvals, colors=colors,
                           balance_alpha=balance_alpha)

    # parameter indices
    parameters_ind = list(range(1, xs.shape[1] + 1))[::-1]

    # plot parameters
    ax.xaxis.set_major_locator(MaxNLocator(integer=True))
    for j_x, x in reversed(list(enumerate(xs))):
        if j_x == 0:
            tmp_legend = legend_text
        else:
            tmp_legend = None
        ax.plot(x, parameters_ind,
                linestyle,
                color=colors[j_x],
                marker='o',
                label=tmp_legend)

    ax.set_yticks(parameters_ind)
    if x_labels is not None:
        ax.set_yticklabels(x_labels)

    # draw bounds
    parameters_ind = np.array(parameters_ind).flatten()
    if lb is not None:
        lb = np.array(lb, dtype="float64")
        ax.plot(lb.flatten(), parameters_ind, 'k--', marker='+')
    if ub is not None:
        ub = np.array(ub, dtype="float64")
        ax.plot(ub.flatten(), parameters_ind, 'k--', marker='+')

    ax.set_xlabel('Parameter value')
    ax.set_ylabel('Parameter')
    ax.set_title('Estimated parameters')
    if legend_text is not None:
        ax.legend()

    return ax


def handle_inputs(
        result: Result, parameter_indices: List[int],
        lb: Optional[Union[np.ndarray, List[float]]] = None,
        ub: Optional[Union[np.ndarray, List[float]]] = None,
        start_indices: Optional[Union[int, Iterable[int]]] = None
) -> Tuple[np.ndarray, np.ndarray, List[str], np.ndarray, List[np.ndarray]]:
    """
    Computes the correct bounds for the parameter indices to be plotted and
    outputs the corresponding parameters and their labels

    Parameters
    ----------

    result:
        Optimization result obtained by 'optimize.py'.

    parameter_indices:
        Specifies which parameters should be plotted.

    lb, ub:
        If not None, override result.problem.lb, problem.problem.ub.
        Dimension either result.problem.dim or result.problem.dim_full.

    start_indices:
        list of integers specifying the multistarts to be plotted or
        int specifying up to which start index should be plotted

    Returns
    -------

    lb, ub:
        Dimension either result.problem.dim or result.problem.dim_full.

    x_labels:
        ytick labels to be applied later on

    fvals:
        objective function values which are needed for plotting later

    xs:
        parameter values which will be plotted later
    """

    # retrieve results
    fvals = result.optimize_result.get_for_key('fval')
    xs = result.optimize_result.get_for_key('x')

    # parse indices which should be plotted
    if start_indices is not None:
        start_indices = process_start_indices(start_indices, len(fvals))

        # reduce number of displayed results
        xs_out = [xs[ind] for ind in start_indices]
        fvals_out = [fvals[ind] for ind in start_indices]
    else:
        # use non-reduced versions
        xs_out = xs
        fvals_out = fvals

    # get bounds
    if lb is None:
        lb = result.problem.lb_full
    if ub is None:
        ub = result.problem.ub_full

    # get labels
    x_labels = result.problem.x_names

    # handle fixed and free indices
    if len(parameter_indices) < result.problem.dim_full:
        for ix, x in enumerate(xs_out):
            xs_out[ix] = result.problem.get_reduced_vector(x,
                                                           parameter_indices)
        lb = result.problem.get_reduced_vector(lb, parameter_indices)
        ub = result.problem.get_reduced_vector(ub, parameter_indices)
        x_labels = [x_labels[int(i)] for i in parameter_indices]
    else:
        lb = result.problem.get_full_vector(lb)
        ub = result.problem.get_full_vector(ub)

    return lb, ub, x_labels, fvals_out, xs_out



import h5py
import numpy as np
import pandas as pd
import os
from typing import Callable, Union, Sequence

from .constants import (EnsembleType, OUTPUT, UPPER_BOUND, LOWER_BOUND,
                        PREDICTION_RESULTS, PREDICTION_ID, SUMMARY,
                        OPTIMIZE, SAMPLE, X_NAMES, TIMEPOINTS, OUTPUT_IDS)
from ..predict import PredictionConditionResult, PredictionResult
from pathlib import Path
from .ensemble import (Ensemble, EnsemblePrediction)
from ..store import read_result, get_or_create_group, write_array


def read_from_csv(path: str,
                  sep: str = '\t',
                  index_col: int = 0,
                  headline_parser: Callable = None,
                  ensemble_type: EnsembleType = None,
                  lower_bound: np.ndarray = None,
                  upper_bound: np.ndarray = None):
    """
    function for creating an ensemble from a csv file

    Parameters
    ----------
    path:
        path to csv file to read in parameter ensemble
    sep:
        separator in csv file
    index_col:
        index column in csv file
    headline_parser:
        A function which reads in the headline of the csv file and converts it
        into vector_tags (see constructor of Ensemble for more details)
    ensemble_type:
        Ensemble type: representative sample or random ensemble
    lower_bound:
        array of potential lower bounds for the parameters
    upper_bound:
        array of potential upper bounds for the parameters

    Returns
    -------
    result:
        Ensemble object of parameter vectors
    """
    # get the data from the csv
    ensemble_df = pd.read_csv(path, sep=sep, index_col=index_col)

    # set the type of the ensemble
    if ensemble_type is None:
        ensemble_type = EnsembleType.ensemble

    return read_from_df(dataframe=ensemble_df,
                        headline_parser=headline_parser,
                        ensemble_type=ensemble_type,
                        lower_bound=lower_bound,
                        upper_bound=upper_bound)


def read_ensemble_from_hdf5(filename: str,
                            input_type: str = OPTIMIZE,
                            remove_burn_in: bool = True,
                            chain_slice: slice = None,
                            cutoff: float = np.inf,
                            max_size: int = np.inf
                            ):
    """
    Create an ensemble from an HDF5 storage file.

    Parameters
    ----------
    filename:
        Name or path of the HDF5 file.
    input_type:
        Which type of ensemble to create. From History, from
        Optimization or from Sample.

    Returns:
    -------
    ensemble:
        Ensemble object of parameter vectors
    """
    # TODO: add option HISTORY. Need to fix
    #  reading history from hdf5.
    if input_type == OPTIMIZE:
        result = read_result(filename=filename,
                             optimize=True)
        return Ensemble.from_optimization_endpoints(result=result,
                                                    cutoff=cutoff,
                                                    max_size=max_size)
    elif input_type == SAMPLE:
        result = read_result(filename=filename,
                             sample=True)
        return Ensemble.from_sample(result=result,
                                    remove_burn_in=remove_burn_in,
                                    chain_slice=chain_slice)
    else:
        raise ValueError('The type you provided was neither '
                         f'"{SAMPLE}" nor "{OPTIMIZE}". Those are '
                         'currently the only supported types. '
                         'Please choose one of them.')


def read_from_df(dataframe: pd.DataFrame,
                 headline_parser: Callable = None,
                 ensemble_type: EnsembleType = None,
                 lower_bound: np.ndarray = None,
                 upper_bound: np.ndarray = None):
    """
    function for creating an ensemble from a csv file

    Parameters
    ----------
    dataframe:
        pandas.DataFrame to read in parameter ensemble
    headline_parser:
        A function which reads in the headline of the csv file and converts it
        into vector_tags (see constructor of Ensemble for more details)
    ensemble_type:
        Ensemble type: representative sample or random ensemble
    lower_bound:
        array of potential lower bounds for the parameters
    upper_bound:
        array of potential upper bounds for the parameters

    Returns
    -------
    result:
        Ensemble object of parameter vectors
    """
    # if we have a parser to make vector_tags from column names, we use it
    vector_tags = None
    if headline_parser is not None:
        vector_tags = headline_parser(list(dataframe.columns))
    # set the type of the ensemble
    if ensemble_type is None:
        ensemble_type = EnsembleType.ensemble

    return Ensemble(x_vectors=dataframe.values,
                    x_names=list(dataframe.index),
                    vector_tags=vector_tags,
                    ensemble_type=ensemble_type,
                    lower_bound=lower_bound,
                    upper_bound=upper_bound)


def write_ensemble_prediction_to_h5(ensemble_prediction: EnsemblePrediction,
                                    output_file: str,
                                    base_path: str = None):
    # parse base path
    base = Path('')
    if base_path is not None:
        base = Path(base_path)

    # open file
    with h5py.File(output_file, 'a') as f:
        # write prediction ID if available
        if ensemble_prediction.prediction_id is not None:
            f.create_dataset(os.path.join(base, PREDICTION_ID),
                             data=ensemble_prediction.prediction_id)

        # write lower bounds per condition, if available
        if ensemble_prediction.lower_bound is not None:
            if isinstance(ensemble_prediction.lower_bound[0], np.ndarray):
                lb_grp = get_or_create_group(f, LOWER_BOUND)
                for i_cond, lower_bounds in \
                        enumerate(ensemble_prediction.lower_bound):
                    condition_id = (
                        ensemble_prediction
                        .prediction_results[0]
                        .condition_ids[i_cond]
                    )
                    write_array(lb_grp, condition_id, lower_bounds)
            elif isinstance(ensemble_prediction.lower_bound[0], float):
                f.create_dataset(LOWER_BOUND,
                                 data=ensemble_prediction.lower_bound)

        # write upper bounds per condition, if available
        if ensemble_prediction.upper_bound is not None:
            if isinstance(ensemble_prediction.upper_bound[0], np.ndarray):
                ub_grp = get_or_create_group(f, UPPER_BOUND)
                for i_cond, upper_bounds in \
                        enumerate(ensemble_prediction.upper_bound):
                    condition_id = \
                        ensemble_prediction.prediction_results[
                            0].condition_ids[i_cond]
                    write_array(ub_grp, condition_id, upper_bounds)
            elif isinstance(ensemble_prediction.upper_bound[0], float):
                f.create_dataset(UPPER_BOUND,
                                 data=ensemble_prediction.upper_bound)

        # write summary statistics to h5 file
        for summary_id, summary in \
                ensemble_prediction.prediction_summary.items():
            if summary is None:
                continue
            tmp_base_path = os.path.join(base, f'{SUMMARY}_{summary_id}')
            f.create_group(tmp_base_path)
            summary.write_to_h5(output_file, base_path=tmp_base_path)

        # write the single prediction results
        for i_result, result in \
                enumerate(ensemble_prediction.prediction_results):
            tmp_base_path = os.path.join(base,
                                         f'{PREDICTION_RESULTS}_{i_result}')
            result.write_to_h5(output_file, base_path=tmp_base_path)


def get_prediction_dataset(ens: Union[Ensemble, EnsemblePrediction],
                           prediction_index: int = 0) -> np.ndarray:
    """
    Extract an array of prediction from either an Ensemble object which
    contains a list of predictions of from an EnsemblePrediction object.

    Parameters
    ==========
    ens:
        Ensemble objects containing a set of parameter vectors and a set of
        predictions or EnsemblePrediction object containing only predictions

    prediction_index:
        index telling which prediction from the list should be analyzed

    Returns
    =======
    dataset:
        numpy array containing the ensemble predictions
    """

    if isinstance(ens, Ensemble):
        dataset = ens.predictions[prediction_index]
    elif isinstance(ens, EnsemblePrediction):
        ens.condense_to_arrays()
        dataset = ens.prediction_arrays[OUTPUT].transpose()
    else:
        raise Exception('Need either an Ensemble object with predictions or '
                        'an EnsemblePrediction object as input. Stopping.')

    return dataset


def read_ensemble_prediction_from_h5(
        predictor: Union[Callable[[Sequence], PredictionResult], None],
        input_file: str):

    # open file
    with h5py.File(input_file, 'r') as f:
        pred_res_list = []
        bounds = {}
        for key in f.keys():
            if key == PREDICTION_ID:
                prediction_id = f[key][()].decode()
                continue
            if key in {LOWER_BOUND, UPPER_BOUND}:
                if isinstance(f[key], h5py._hl.dataset.Dataset):
                    bounds[key] = f[key][:]
                    continue
                bounds[key] = [f[f'{key}/{cond}'][()]
                               for cond in f[key].keys()]
                bounds[key] = np.array(bounds[key])
                continue
            x_names = decode_array(f[f'{key}/{X_NAMES}'][()])
            condition_ids = np.array(decode_array(
                f[f'{key}/condition_ids'][()]
            ))
            pred_cond_res_list = []
            for id, _ in enumerate(condition_ids):
                output = f[f'{key}/{id}/{OUTPUT}'][:]
                output_ids = decode_array(f[f'{key}/{id}/{OUTPUT_IDS}'][:])
                timepoints = f[f'{key}/{id}/{TIMEPOINTS}'][:]
                pred_cond_res_list.append(PredictionConditionResult(
                    timepoints=timepoints,
                    output_ids=output_ids,
                    output=output,
                    x_names=x_names
                ))
            pred_res_list.append(PredictionResult(
                conditions=pred_cond_res_list,
                condition_ids=condition_ids
            ))
        return EnsemblePrediction(predictor=predictor,
                                  prediction_id=prediction_id,
                                  prediction_results=pred_res_list,
                                  )


def decode_array(array: np.ndarray) -> np.ndarray:
    """Decodes array of bytes to string"""
    for i in range(len(array)):
        array[i] = array[i].decode()
    return array
